import sys
#sys.path.insert(0, '/mumpce/')
import CheKiPEUQ.mumpce.Project as mumpceProject
import CheKiPEUQ.mumpce.solution as mumpceSolution
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import cm #EAW 2020/01/07
import copy

class plotting_functions_class():
    def __init__(self, UserInput, samples = False): # The plots require samples.  Other plot settings are probably plotting-package specific.
        self.UserInput = UserInput
#        if not samples: #Deprecated.
#            print("Warning: Pass in the 'samples' keyword argument containing a numpy array of samples to plot.")
    
    def mu_and_cov_from_samples(self):
        mu = np.mean(self.samples, axis = 0)
        cov = np.cov(self.samples,rowvar=False)
        return mu, cov

    def mumpce_plots(self, model_parameter_info = {}, active_parameters = [], pairs_of_parameter_indices = [], posterior_mu_vector = 0, posterior_cov_matrix = 0, prior_mu_vector = 0, prior_cov_matrix = 0, contour_settings_custom = {'figure_name','fontsize','num_y_ticks','num_x_ticks','colormap_posterior_customized','colormap_prior_customized','contours_normalized','colorbars','axis_limits'}): # Pass empty keyword arguments for important parameters.  That way, warnings may be issued if they are not set.  There is not really a good default for these keyword arguments.  They depend entirely on the nature of the data being plotted.
        mumpceProjectObject = mumpceProject.Project() # A mumpce project object must be created.
        if len(model_parameter_info) == 0:
            print("Pass the 'model_parameter_info' argument to the mumpce_plots function.")
            model_parameter_info = np.array([{'parameter_number': 0, 'parameter_name': 'Parameter 0'},{'parameter_number': 1, 'parameter_name': 'Parameter 1'}])
        if len(active_parameters) == 0:
            print("Pass the 'active_parameters' argument to the mumpce_plots function.")
            active_parameters = np.array([0, 1]) 
        mumpceProjectObject.active_parameters = active_parameters
        #if len(pairs_of_parameter_indices) == 0:
        #    print("Pass the 'pairs_of_parameter_indices' argument to the mumpce_plots function.")
        #    mumpceProjectObject.pairsOfParameterIndices = [[0, 1]]
        #else:
        mumpceProjectObject.pairsOfParameterIndices = pairs_of_parameter_indices
        #if not posterior_mu_vector == 0:
        #    print("Pass the 'posterior_mu_vector' argument to the mumpce_plots function.")
        #    posterior_mu_vector = np.array([-0.58888733,1.1200355])
        #if not posterior_cov_matrix == 0:
        #    print("Pass the 'posterior_cov_matrix' argument to the mumpce_plots function.")
        #    posterior_cov_matrix = np.array([[ 0.0148872,-0.01894579],[-0.01894579,0.04284732]])
        #if not prior_mu_vector == 0:
        #    print("Pass the 'prior_mu_vector' argument to the mumpce_plots functions.")
        #    prior_cov_matrix = np.array([-0.98888733,0.8200355])
        #if not prior_cov_matrix == 0:
        #    prior_cov_matrix = 10*posterior_cov_matrix
        #    print("Pass the 'prior_cov_matrix' argument to the mumpce_plots functions.")
        mumpceProjectObject.model_parameter_info = model_parameter_info
        mumpceSolutionsObject = mumpceSolution.Solution(posterior_mu_vector, posterior_cov_matrix, initial_x=prior_mu_vector, initial_covariance=prior_cov_matrix)
        mumpceProjectObject.solution = mumpceSolutionsObject
        #if hasattr(UserInput,'figure_name'):
        #    contour_settings_custom['figure_name']=UserInput.figure_name
        #else:
        #    contour_settings_custom['figure_name']='mumpce_plots_04'
        #if hasattr(UserInput,'fontsize'):
        #    contour_settings_custom['fontsize'] = UserInput.fontsize        
        #else:
        #    contour_settings_custom['fontsize'] = 'auto'
        #if hasattr(UserInput,'num_y_ticks'):
        #    contour_settings_custom['num_y_ticks'] = UserInput.num_y_ticks
        #else:
        #    contour_settings_custom['num_y_ticks'] = 'auto'
        #if hasattr(UserInput,'num_x_ticks'):
        #    contour_settings_custom['num_x_ticks'] = UserInput.num_x_ticks
        #else:
        #    contour_settings_custom['num_x_ticks'] = 'auto'
        #if hasattr(UserInput,'colormap_posterior_customized'):
        #    contour_settings_custom['colormap_posterior_customized'] = UserInput.colormap_posterior_customized
        #else:
        #    contour_settings_custom['colormap_posterior_customized'] = "Oranges"
        #if hasattr(UserInput,'colormap_prior_customized'):
        #    contour_settings_custom['colormap_prior_customized'] = UserInput.colormap_prior_customized
        #else:
        #    contour_settings_custom['colormap_prior_customized'] = "Greens"
        #if hasattr(UserInput,'contours_normalized'):
        #    contour_settings_custom['contours_normalized'] = UserInput.contours_normalized
        #else:
        #    contour_settings_custom['contours_normalized'] = False
        #if hasattr(UserInput,'center_on'):
        #    contour_settings_custom['center_on'] = UserInput.center_on
        #else:
        #    contour_settings_custom['center_on'] = 'prior'
        #if hasattr(UserInput,'colorbars'):
        #    contour_settings_custom['colorbars'] = UserInput.colorbars
        #else:
        #    contour_settings_custom['colorbars'] = True
        mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)


    def seaborn_scatterplot_matrix(self):
        #fig0, ax0 = plt.subplots()
        #if UserInput.verbose:
        #    print(np.mean(rate_tot_array,axis = 0))
        #ax0.plot(np.array(experiments_df['AcH - T']),np.mean(rate_tot_array,axis = 0), 'r')
        #ax0.plot(np.array(experiments_df['AcH - T']),np.array(experiments_df['AcHBackgroundSubtracted'])/2000,'g')
        #ax0.set_ylim([0.00, 0.025])
        #ax0.set_xlabel('T (K)')
        #ax0.set_ylabel(r'$rate (s^{-1})$')
        #ax0.legend(['model posterior', 'experiments'])
        #fig0.tight_layout()
        #fig0.savefig('tprposterior.png', dpi=220)
        #posterior_df = pd.DataFrame(samples,columns=[UserInput.parameterNamesAndMathTypeExpressionsDict[x] for x in UserInput.parameterNamesList])
        #pd.plotting.scatter_matrix(posterior_df)
        #plt.savefig('scatter_matrix_posterior.png',dpi=220)
        return

    def rate_tot_plot(self):
        return

def sampledParameterHistogramMaker(parameterSamples, parameterName,parameterNamesAndMathTypeExpressionsDict, sampledParameterFiguresDictionary, sampledParameterAxesDictionary):
        parameterIndex = list(parameterNamesAndMathTypeExpressionsDict).index(parameterName)
        sampledParameterFiguresDictionary[parameterName], sampledParameterAxesDictionary[parameterName] = plt.subplots()   #making plt objects    
        sampledParameterAxesDictionary[parameterName].hist(parameterSamples[:,parameterIndex]) #filling the object with data
        #setting the labels etc. and then exporting.
        axis_font = {'size':'16'} #TODO: Make this size a setting that can be changed. #Code was made following answer by "binaryfunt" here https://stackoverflow.com/questions/3899980/how-to-change-the-font-size-on-a-matplotlib-plot
        sampledParameterAxesDictionary[parameterName].set_ylabel('frequency', **axis_font)
        sampledParameterAxesDictionary[parameterName].set_xlabel(parameterNamesAndMathTypeExpressionsDict[parameterName], **axis_font)

        sampledParameterAxesDictionary[parameterName].tick_params(axis='x', labelsize=16) #TODO: make these labels sizes a setting that can be changed.
        sampledParameterAxesDictionary[parameterName].tick_params(axis='y', labelsize=16)
        sampledParameterFiguresDictionary[parameterName].tight_layout()
        sampledParameterFiguresDictionary[parameterName].savefig('Histogram_sampling_'+parameterName+'.png', dpi=220)

        
        
        #The above block makes code kind of like this in a dynamic fashion. Since we know how many we will need, a dictionary is used to avoid the need for 'exec' statements when making new parameters.
        # fig2, ax2 = plt.subplots()
        # ax2.hist(samples[:,1])
        # ax2.set_ylabel('frequency')
        # ax2.set_xlabel(r'$E_{a2}$')
        # fig2.tight_layout()
        # fig2.savefig('Ea2.png', dpi=220)

    #Make histograms for each parameter. Need to make some dictionaries where relevant objects will be stored.
def makeHistogramsForEachParameter(parameterSamples,parameterNamesAndMathTypeExpressionsDict):
    sampledParameterFiguresDictionary = copy.deepcopy(parameterNamesAndMathTypeExpressionsDict) #This must be a deep copy to perserve original.
    sampledParameterAxesDictionary = copy.deepcopy(parameterNamesAndMathTypeExpressionsDict) #This must be a deep copy to preserve original.
    for key in parameterNamesAndMathTypeExpressionsDict:
        parameterName = key
        sampledParameterHistogramMaker(parameterSamples, parameterName,parameterNamesAndMathTypeExpressionsDict, sampledParameterFiguresDictionary, sampledParameterAxesDictionary)        

def createSimulatedResponsesPlot(x_values, listOfYArrays, plot_settings={}, listOfYUncertaintiesArrays=[], showFigure=True):
    exportFigure = True #This variable should be moved to an argument or something in plot_settings.
    #First put some defaults in if not already defined.
    x_values = np.array(x_values)
    if len(np.shape(x_values)) == 0: #This 2d line breaks regular arrays, but works when a 'zero length' array comes in (I don't understand how a zero length array can occur, but it has happened.)
        x_values = np.atleast_2d(x_values)
    if 'x_label' not in plot_settings: plot_settings['x_label'] = ''
    if 'y_label' not in plot_settings: plot_settings['y_label'] = ''
    if 'legendLabels' not in plot_settings: plot_settings['legendLabels'] = ''
    if 'figure_name' not in plot_settings: plot_settings['figure_name'] = 'simulatedResponsesPlot'
    if 'dpi' not in plot_settings: plot_settings['dpi']=220
    fig0, ax0 = plt.subplots()
    if 'fontdict' in plot_settings: 
        #There are various things that could be added to this fontdict. #https://www.tutorialexample.com/understand-matplotlib-fontdict-a-beginner-guide-matplotlib-tutorial/
        fontdict = plot_settings['fontdict']
        if 'size' in fontdict:
            ax0.tick_params(axis='x', labelsize=fontdict['size'])
            ax0.tick_params(axis='y', labelsize=fontdict['size'])
    else:
        fontdict = None #initializing with the matplotlib default
    ax0.set_xlabel(plot_settings['x_label'], fontdict=fontdict)
    ax0.set_ylabel(plot_settings['y_label'], fontdict=fontdict) #TODO: THis is not yet generalized (will be a function)
    #The error linewidth is going to be set to a thick value if we have a small number of points.
    if 'error_linewidth' in plot_settings: error_linewidth = plot_settings['error_linewidth'] # making a convenient local variable.
    if str(error_linewidth).lower() == 'auto':
        if len(x_values) == 1:
            error_linewidth = 20
        elif len(x_values) > 10:
            error_linewidth = 1
        elif len(x_values) <= 10:
            error_linewidth = 10
    if str(error_linewidth).lower() == 'none': error_linewidth = 0 #This will hide the rror bars if they are not desired.
    if 'y_range' in plot_settings: ax0.set_ylim(plot_settings['y_range'] )
    if len(listOfYArrays) == 3: #This generally means observed, mu_guess, map, in that order.
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):           
                ax0.plot(x_values,listOfYArrays[0],'g')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line..
                ax0.plot(x_values,listOfYArrays[1], '#00A5DF')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
                ax0.plot(x_values,listOfYArrays[2], 'r') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
                    
        if len(x_values) == 1: #This means there are single points, and we need to make symbols, by adding an "o".
            for seriesIndex in range(len(listOfYArrays)):           
                ax0.plot(x_values,listOfYArrays[0],'go')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='o', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], 'co')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[2], 'ro') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
    elif len(listOfYArrays) == 4: #This generally means observed, mu_guess, map, mu_app
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values,listOfYArrays[0],'g')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], '#00A5DF')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                                             
                ax0.plot(x_values,listOfYArrays[2], 'r') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[3], 'k')  #k is black.
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[3], yerr=np.array(listOfYUncertaintiesArrays[3]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
        if len(x_values) == 1: #This means there are single points, and we need to make symbols, by adding an "o".
                ax0.plot(x_values,listOfYArrays[0],'go')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], 'co')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[2], 'ro') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[3], 'ko')  #k is black. https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[3]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
    else:
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values[0],listOfYArrays[seriesIndex])
        if len(x_values) == 1: #This means there are single points, and we need to make symbols.
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values[0],listOfYArrays[seriesIndex], 'o')            
    if plot_settings['legend'] == True:
        ax0.legend(plot_settings['legendLabels']) #legends must be after plots are made.
    fig0.tight_layout()
    if exportFigure==True:
        fig0.savefig(plot_settings['figure_name'] + '.png', dpi=plot_settings['dpi'])
    if showFigure==False:
        plt.close(fig0)
    return fig0

def makeTrisurfacePlot(xValues, yValues, zValues, exportFigure = True, figure_name="TrisurfacePlot", showFigure=True):
    fig1, ax1 =  plt.subplots(1)
    ax1 = plt.axes(projection ='3d')
    image = ax1.plot_trisurf(xValues,yValues, zValues)
    if exportFigure == True:
        fig1.savefig(figure_name + '.png')
    if showFigure==False:
        plt.close(fig1)
    return fig1, ax1, image

def makeMeshGridSurfacePlot(XX, YY, ZZ,  plot_settings = {}, exportFigure = True, figure_name="MeshGridSurfacePlot", showFigure=True):
    #TODO: plot_settings should be used for axis labels etc, like above.
    #TODO: create a UserInput variable named info_gain_plot_settings (like what the other cases have).
    fig1,ax1 = plt.subplots(figsize=(5,5))
    #ax = fig.add_subplot(111, projection='3d')
    surf = ax1.pcolor(XX,YY,ZZ,cmap=matplotlib.cm.coolwarm)
    # ax1.set_xlabel('Temperature (K)')
    # ax1.set_ylabel(r'$p_A$')
    #ax.set_zlabel('Information Gain')
    #ax1.set_xticks(temperatures)
    #ax1.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
    ax1.set_title('Information Gain Surface')
    fig1.colorbar(surf, shrink=0.5, aspect=5)
    if exportFigure==True:
        fig1.savefig(figure_name + '.png')
    if showFigure==False:
        plt.close(fig1)
    return fig1, ax1#, image
    


try:
    from mpi4py import MPI
    rank = MPI.COMM_WORLD.Get_rank()
    numProcessors = MPI.COMM_WORLD.Get_size()
    numSimulations = numProcessors - 1 #Normally, we will use rank 0 for controlling.
    currentProcessorNumber = rank
    #now this number can be accessed from elsewhere using parallel_processing.CurrentProcessorNumber
    if numProcessors > 1:
        using_mpi = True
    else:
        using_mpi = True
        
except:
    currentProcessorNumber=0
    numProcessors=1
    using_mpi = False






import cantera as ct
import cantera.ck2cti as ck2cti
import canteraSimulate
import numpy as np
import canteraKineticsParametersParser 


def main():
    #Here is an example usage. The things that must be defined in advance are...
    # a) a model name.
    # b) The simulation settings are set in a python file which then becomes imported and an argument. This can then be changed by a script.
    # c) The reactions_parameters_array must be fed as an array or as a file which contains such an array with a header. This includes the reaction equations and their Arrhenius initial values.
    # d) The cti_top_info_string. Right now, this file is made 'manually'.
    
    print("####EXAMPLE 1#####")
    #Here is an example that leaves the settings as they are in the settings file. 
    #For the cti_stop_info_string the setting "None" is used, but that just means the code will automatically look for it in a file. Though it could be provided instead.
    import ceO2_input_simulation_settings #The user may change settings in the python file with the same name.
    ceO2_input_simulation_settings.print_frequency = 100 #This makes the simulation not print things out during checkpoints.
    canteraSimulate.create_cti_and_SimulatePFRorTPRwithCantera("ceO2", "ceO2_input_reactions_parameters.csv", ceO2_input_simulation_settings, cti_top_info_string = None, write_cti_to_file = True)
    #Note that one can change from static to PFR by changing a variable inside ceO2_simulation_settings. 
    #To get the output as python objects and not only as file exports, see example 2.

    print("####EXAMPLE 2#####")
    #Here is an example useage with a loop to scan different parameter settings.
    import ceO2_input_simulation_settings #The user may change settings in the python file with the same name.
    model_name = "ceO2"
    reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    import copy
    modified_reactions_parameters_array = copy.deepcopy(reactions_parameters_array)
    #This time we are going to get the cti_top_info string so that we don't waste time reading it repeatedly.        
    cti_top_info_filename = model_name + "_cti_top_info.cti"
    with open(cti_top_info_filename, "r") as cti_top_info_file:
        cti_top_info_string = cti_top_info_file.read()     
    #Now we execute a loop with some changes...
    for multiplyingFactor in range(1,11):
        ceO2_input_simulation_settings.print_frequency = None #This makes the simulation not print things out during checkpoints.
        ceO2_input_simulation_settings.exportOutputs = True #This will be turned off during a paramter optimization.
        modified_reactions_parameters_array[1][5] = str(float(reactions_parameters_array[1][5])*(1+multiplyingFactor/10)) #This modifies the desorption rate constant. #note that the kinetic parameter starts as a string.
        concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
        canteraSimulate.create_cti_and_SimulatePFRorTPRwithCantera("ceO2", modified_reactions_parameters_array, ceO2_input_simulation_settings, cti_top_info_string = cti_top_info_string)
        np.savetxt("ceO2_output_rates_all_"+str(multiplyingFactor)+".csv", rates_all_array, delimiter=",", comments='', header=rates_all_array_header)
        print("round", multiplyingFactor, "Simulation Finished")
        
    print("####EXAMPLE 3#####")
    #This is like example 2, only now we use the modifyAllAdjustableReactionParametersArrayValues function rather than sticking values in by simple index.
    model_name = "ceO2"
    import ceO2_input_simulation_settings #The user may change settings in the python file with the same name.
    reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    
    #Let's get the modfiable parameter indices and the original values before any modifications...
    modifiableParameterIndices = canteraKineticsParametersParser.findModifiableParameterIndices(reactions_parameters_array)
    originalModifiableReactionParametersValues = canteraKineticsParametersParser.getAllModifiableReactionParametersValues(reactions_parameters_array, modifiableParameterIndices)
    print(modifiableParameterIndices)
    print(originalModifiableReactionParametersValues)
    #The original parameter values are:
    # ['1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '67400.0', '0.0', '0.0', '-6000.000000000001']
    
    #This time we are again going to get the cti_top_info string so that we don't waste time reading it repeatedly.        
    cti_top_info_filename = model_name + "_cti_top_info.cti"
    with open(cti_top_info_filename, "r") as cti_top_info_file:
        cti_top_info_string = cti_top_info_file.read()     
    #Now we execute a loop with some changes... but this time we're going to inject adjustedModifiableParameters
    adjustedModifiableParametersSamplings = [
    ['1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '67400.0', '0.0', '0.0', '-6000.000000000001', '1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '67400.0', '0.0', '0.0', '-6000.000000000001'], #First one is actually not modified.
    ['1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '87400.0', '0.0', '0.0', '-6000.000000000001', '1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '67400.0', '0.0', '0.0', '-6000.000000000001'],
    ['1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '107400.0', '0.0', '0.0', '-6000.000000000001', '1.0', '0.0', '2000.0', '10000000000000.0', '0.0', '67400.0', '0.0', '0.0', '-6000.000000000001']
    ] #IT IS IMPORTANT TO REALIZE THAT THESE ARE NOT THE FULL MOIDFIED PARAMETER ARRAYS, DON'T GET THE TWO CONFUSED!
    
    for caseIndex, adjustedModifiableParametersCase in enumerate(adjustedModifiableParametersSamplings):
        ceO2_input_simulation_settings.print_frequency = None #This makes the simulation not print things out during checkpoints.
        ceO2_input_simulation_settings.exportOutputs = True #This will be turned off during a paramter optimization.
        modified_reactions_parameters_array = canteraKineticsParametersParser.modifyAllAdjustableReactionParametersArrayValues(reactions_parameters_array, adjustedModifiableParametersCase)
        concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
        canteraSimulate.create_cti_and_SimulatePFRorTPRwithCantera("ceO2", modified_reactions_parameters_array, ceO2_input_simulation_settings, cti_top_info_string = cti_top_info_string)
        np.savetxt("ceO2_output_rates_all_"+"FullCTIsamplingCase"+str(caseIndex)+".csv", rates_all_array, delimiter=",", comments='', header=rates_all_array_header)              
        print("sampling case FullCTIsamplingCase", caseIndex, "Simulation Finished")

    print("####EXAMPLE 4#####")
    #This is like example 3, only now we use the modifyReactionsInOnePhase rather than full cti.
    model_name = "ceO2"
    import ceO2_input_simulation_settings #The user may change settings in the python file with the same name.
    reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    
    #This time we are again going to get the cti_top_info string so that we don't waste time reading it repeatedly.        
    cti_top_info_filename = model_name + "_cti_top_info.cti"
    with open(cti_top_info_filename, "r") as cti_top_info_file:
        cti_top_info_string = cti_top_info_file.read()     
    #Now we execute a loop with some changes... but this time we're going to inject entire already modified parameter arrays.
    adjustedModifiableParametersSamplings = [
            [
            ['0001','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde1-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0002','surface_reaction','Acetaldehyde1-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','67400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],
            ['0003','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde2-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0004','surface_reaction','Acetaldehyde2-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','67400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],
            ],#First one is actually not modified.
    
            [
            ['0001','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde1-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0002','surface_reaction','Acetaldehyde1-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','87400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],
            ['0003','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde2-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0004','surface_reaction','Acetaldehyde2-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','67400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],            
            ],
            
            
            [
            ['0001','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde1-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0002','surface_reaction','Acetaldehyde1-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','107400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],
            ['0003','surface_reaction','Acetaldehyde + CeCation(S) => Acetaldehyde2-Ce(S)','1.0','0.0','2000.0','nan','nan','nan','None','True'],
            ['0004','surface_reaction','Acetaldehyde2-Ce(S) => Acetaldehyde + CeCation(S)','10000000000000.0','0.0','67400.0','0.0','0.0','-6000.000000000001','Acetaldehyde(S)','False'],            
            ]
    ] 
    
    for caseIndex, adjustedModifiableParametersCase in enumerate(adjustedModifiableParametersSamplings):
        ceO2_input_simulation_settings.print_frequency = None #This makes the simulation not print things out during checkpoints.
        ceO2_input_simulation_settings.exportOutputs = True #This will be turned off during a paramter optimization.
        concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
        canteraSimulate.modify_reactions_and_SimulatePFRorTPRwithCantera(model_name, adjustedModifiableParametersCase, ceO2_input_simulation_settings, canteraPhases=canteraPhases)        
        np.savetxt("ceO2_output_rates_all_"+"ModifyReactionssamplingCase"+str(caseIndex)+".csv", rates_all_array, delimiter=",", comments='', header=rates_all_array_header)
        print("sampling case ModifyReactionssamplingCase", caseIndex, "Simulation Finished")
        
    print("####EXAMPLE 5b#####")
    #This is like example 4, only now we are going to use the applyPieceWiseCoverageDependence feature, which requires us to fill ceO2_input_simulation_settings.piecewise_coverage_dependences
    model_name = "ceO2"
    import ceO2_input_simulation_settings #The user may change settings in the python file with the same name.
    reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    #We are going to set the starting coverage of "Acetaldehyde1-Ce(S)"  to 1 and the rest to 0 so that it's effectively a 1 species PE-TPR experiment.
    ceO2_input_simulation_settings.surface_coverages = 'CeCation(S):0 Acetaldehyde1-Ce(S):1 Acetaldehyde2-Ce(S):0 OAnion(S):0'
        
    #This time we are again going to get the cti_top_info string so that we don't waste time reading it repeatedly.        
    cti_top_info_filename = model_name + "_cti_top_info.cti"
    with open(cti_top_info_filename, "r") as cti_top_info_file:
        cti_top_info_string = cti_top_info_file.read()     
        
    ceO2_input_simulation_settings.piecewise_coverage_dependence = True
    ceO2_input_simulation_settings.original_reactions_parameters_array = reactions_parameters_array
    piecewise_coverage_intervals = np.array([0,0.1,0.2,0.3,0.4,0.5,1.0])
    species_name = "Acetaldehyde1-Ce(S)" #In this example we make the parameters depend only on the coverage of species Acetaldehyde1-Ce(S).
    kineticParameterName = "A"
    modifiers_A = ( [0,0,0,0,0,0,0], [-1,-1,-1,-1,-1,-1,0], [0,0,0,0,0,0,0], [0,0,0,0,0,0,0])
    canteraKineticsParametersParser.populatePiecewiseCoverageDependence(ceO2_input_simulation_settings, reactions_parameters_array, species_name, kineticParameterName, piecewise_coverage_intervals, modifiers_A )
    kineticParameterName = "E" 
    modifiers_E = ( [60000,50000,40000,30000,20000,10000,0], [60000,50000,40000,30000,20000,10000,0], [0,0,0,0,0,0,0], [60000,50000,40000,30000,20000,10000,0])
    
    #Now we call a helper function to make sure the activation energy is decreasing with coverage. It returns true if the final activation energy will decrease with coverage.
    checkResult = canteraKineticsParametersParser.descendingLinearEWithPiecewiseOffsetCheckOneReactionAllReactions(reactions_parameters_array=reactions_parameters_array, piecewise_coverage_intervals_all_reactions=piecewise_coverage_intervals, E_offsets_array_all_reactions=modifiers_E, verbose = False)
    if checkResult == True: #In this example, we are only proceeding if the coverage dependance is always decreasing with coverage.
        canteraKineticsParametersParser.populatePiecewiseCoverageDependence(ceO2_input_simulation_settings, reactions_parameters_array, species_name, kineticParameterName, piecewise_coverage_intervals, modifiers_E)
        #Now simulate!
        concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
        canteraSimulate.modify_reactions_and_SimulatePFRorTPRwithCantera(model_name, adjustedModifiableParametersCase, ceO2_input_simulation_settings, canteraPhases=canteraPhases)
        np.savetxt("ceO2_output_rates_all_"+"Piecewise_coverage_Dependence"+".csv", rates_all_array, delimiter=",", comments='', header=rates_all_array_header)
        print("Piecewise_coverage_Dependence", "Simulation Finished")

if __name__ == '__main__':
    main()    


#Try to do things a bit like https://github.com/python/cpython/blob/master/Lib/collections/__init__.py
#Except instead of putting things here directly in __init__, we'll impor them so they are accessible by importing the module.
from CheKiPEUQ.InverseProblem import *
from CheKiPEUQ.plotting_functions import *
from CheKiPEUQ.mumpce import *
#from InverseProblem import parameter_estimation


import math
import itertools
import numpy
import copy

def RestrictionsPassed(CombinationToCheck,):
	FailedStatus = 0  # This means it has not yet failed.  If any of the checks fail, then we change this to 1.
	# Note: the easy way to relate parameters to k is the following: take the k number and multiply it by 2. That's the parameter for Ea. Subtract 1 and that's the parameter for the pre-exponential.
	# Then subract 1 from each of those to convert it to arrays!
	# if (CombinationToCheck[3-1])*math.exp(-CombinationToCheck[4-1]/(R*T)) < (CombinationToCheck[5-1])*math.exp(-CombinationToCheck[6-1]/(R*T)): FailedStatus=1  #require k2 > k3, so fails if k2 < k3
	# if (CombinationToCheck[3-1])*math.exp(-CombinationToCheck[4-1]/(R*T)) < (CombinationToCheck[7-1])*math.exp(-CombinationToCheck[8-1]/(R*T)): FailedStatus=2 #fail if k2 < k4
	# if (CombinationToCheck[3-1])*math.exp(-CombinationToCheck[4-1]/(R*T)) < (CombinationToCheck[9-1])*math.exp(-CombinationToCheck[10-1]/(R*T)): FailedStatus=3 #fail if k2 < k5
	# if (CombinationToCheck[9-1])*math.exp(-CombinationToCheck[10-1]/(R*T)) <  100*(CombinationToCheck[11-1])*math.exp(-CombinationToCheck[12-1]/(R*T)): FailedStatus=4 #require k5 >> k6, so fails if k5 < 100 * k6
	# if (CombinationToCheck[21-1])*math.exp(-CombinationToCheck[22-1]/(R*T)) < (CombinationToCheck[5-1])*math.exp(-CombinationToCheck[6-1]/(R*T)): FailedStatus=5 #fail if k11 < k3
	# if (CombinationToCheck[21-1])*math.exp(-CombinationToCheck[22-1]/(R*T)) < (CombinationToCheck[7-1])*math.exp(-CombinationToCheck[8-1]/(R*T)): FailedStatus=6 #fail if k11 < k4
	# if (CombinationToCheck[21-1])*math.exp(-CombinationToCheck[22-1]/(R*T)) < (CombinationToCheck[9-1])*math.exp(-CombinationToCheck[10-1]/(R*T)): FailedStatus=7 #fail if k11 < k5
	# if (CombinationToCheck[21-1])*math.exp(-CombinationToCheck[22-1]/(R*T)) < (CombinationToCheck[11-1])*math.exp(-CombinationToCheck[12-1]/(R*T)): FailedStatus=8 #fail if k11 < k6
	# if (CombinationToCheck[13-1])*math.exp(-CombinationToCheck[14-1]/(R*T)) < (CombinationToCheck[5-1])*math.exp(-CombinationToCheck[6-1]/(R*T)): FailedStatus=9 #fail if k7 < k3
	# if (CombinationToCheck[13-1])*math.exp(-CombinationToCheck[14-1]/(R*T)) < (CombinationToCheck[7-1])*math.exp(-CombinationToCheck[8-1]/(R*T)): FailedStatus=10 #fail if k7 < k4
	# if (CombinationToCheck[13-1])*math.exp(-CombinationToCheck[14-1]/(R*T)) > (CombinationToCheck[15-1])*math.exp(-CombinationToCheck[16-1]/(R*T)): FailedStatus=11 #fail if k7 > k8
	# if (CombinationToCheck[13-1])*math.exp(-CombinationToCheck[14-1]/(R*T)) > (CombinationToCheck[17-1])*math.exp(-CombinationToCheck[18-1]/(R*T)): FailedStatus=12 #fail if k7 > k9
	# if (CombinationToCheck[13-1])*math.exp(-CombinationToCheck[14-1]/(R*T)) > (CombinationToCheck[19-1])*math.exp(-CombinationToCheck[20-1]/(R*T)): FailedStatus=13 #fail if k7 > k10
	# if (CombinationToCheck[17-1])*math.exp(-CombinationToCheck[18-1]/(R*T)) < (CombinationToCheck[15-1])*math.exp(-CombinationToCheck[16-1]/(R*T)): FailedStatus=14 #fail if k9 < k8
	# if (CombinationToCheck[19-1])*math.exp(CombinationToCheck[20-1]/(R*T)) < (CombinationToCheck[17-1])*math.exp(-CombinationToCheck[18-1]/(R*T)): FailedStatus=15 #fail if k10 < k9

	# THE FOLLOWING LINES ARE FOR DEBUGGING ETC
	#	print str(CombinationToCheck[9-1]) + ' ' + str(math.exp(-CombinationToCheck[10-1]/(R*T))) + ' ' + str(CombinationToCheck[11-1]) + ' ' + str(math.exp(-CombinationToCheck[12-1]/(R*T)))
	#	print str(CombinationToCheck[9-1]*math.exp(-CombinationToCheck[10-1]/(R*T))) + ' ' + str(CombinationToCheck[11-1]*math.exp(-CombinationToCheck[12-1]/(R*T)))
	#	if CombinationToCheck[9-1]*math.exp(-CombinationToCheck[10-1]/(R*T)) < 100*CombinationToCheck[11-1]*math.exp(-CombinationToCheck[12-1]/(R*T)):
	#		print "true that it failed"
	#	if CombinationToCheck[9-1]*math.exp(-CombinationToCheck[10-1]/(R*T)) > 100*CombinationToCheck[11-1]*math.exp(-CombinationToCheck[12-1]/(R*T)):
	#		print "false that it failed"
	if FailedStatus == 0:
		PassedStatus = 1
	else:
		PassedStatus = 0
	return PassedStatus


def combinationGenerator(OriginalParameters, BaseFactor, NumberOfVariationsInEachDirection,SpreadType="Addition", toFile = False, fileName = "output", headerList = "", numberOfTimesToWriteToFile = 1): #SpreadType can be "Addition" or "Multiplication"
	SpreadType.upper()	#makes userinput uppercase for error protection
	CombinationsSoFar = [OriginalParameters]

	for ParameterP in range(len(OriginalParameters)):  # I am Using array indexing, so if the length returned is "21" that means it will iterate across 22 values since 0 and 21 are both values in loop.
		# Note that we are using "ParameterP" to represent the index of the parameter. We'll use this index to extract parameter values and base multipliers etc..
		# Now that we are looping across each parameter, we will iterate across all CombinationsSoFar, then do the base multiplication on each Combination.
		for CurrentBaseCombinationIndex in range(len(CombinationsSoFar)):  # we can't use "for CurrentBaseCombination in CombinationsSoFar" because that would create an infinite loop due to append.
			# Instead, we use the length and array indexing for the loop, since the length will not be recalculated after the first time.
			CurrentBaseCombination = CombinationsSoFar[CurrentBaseCombinationIndex]
			# Now, for each BaseCombination, we need to do the multiplying in each direction for our current ParameterP index.
			if NumberOfVariationsInEachDirection[ParameterP] > 0:  # We're only going to do this for cases where Variations in each direction is non-zero.
				for i in range(1, NumberOfVariationsInEachDirection[ParameterP] + 1):  # If number of vars is 2 for example, then it'll go "1,2" on the i index.
					# Now for each variation index, we need to make the high and the low.
					# for the low case, it actually requires 2 steps: initialization and then changing the parameter currently being changed.
					NewCombinationLow = list(copy.copy(CurrentBaseCombination))
					if SpreadType == "Multiplication":
						NewCombinationLow[ParameterP] = CurrentBaseCombination[ParameterP] / (BaseFactor[ParameterP] ** i)
					elif SpreadType == "Addition":
						NewCombinationLow[ParameterP] = CurrentBaseCombination[ParameterP] - (BaseFactor[ParameterP] * i)
					RestrictionsPassedLow = RestrictionsPassed(NewCombinationLow)
					if RestrictionsPassedLow == 1:
						CombinationsSoFar.append(tuple(NewCombinationLow))
						#print(CombinationsSoFar)
					elif RestrictionsPassedLow == 0:
						CombinationsSoFar.append(tuple(CurrentBaseCombination))
					# Now for high combinations...
					NewCombinationHigh = list(copy.copy(CurrentBaseCombination))
					if SpreadType == "Multiplication":
						NewCombinationHigh[ParameterP] = CurrentBaseCombination[ParameterP] * (BaseFactor[ParameterP] ** i)
					if SpreadType == "Addition":
						NewCombinationHigh[ParameterP] = CurrentBaseCombination[ParameterP] + (BaseFactor[ParameterP] * i)
					RestrictionsPassedHigh = RestrictionsPassed(NewCombinationHigh)
					if RestrictionsPassedHigh == 1:
						CombinationsSoFar.append(tuple(NewCombinationHigh))
					if RestrictionsPassedHigh == 0:
						CombinationsSoFar.append(tuple(CurrentBaseCombination))

	if toFile == True:
		writeToFile(CombinationsSoFar,headerList, fileName, numberOfTimesToWriteToFile)

	return CombinationsSoFar

def writeToFile(comboGeneratedList, headerList, filename, numberOfTimesToWriteOutput = 1):
	filename+=".csv"
	ListToPrint = [headerList]*1
	ListToPrint.extend(comboGeneratedList)

	#print ListToPrint
	f1 = open(filename, 'w')
	for line in ListToPrint:
		for b in range(1,numberOfTimesToWriteOutput + 1):  # b is an arbitrary counter. we use +1 b/c upper end of range is omitted.
			f1.write(str(line)[1:-1] + '\n')
	f1.close()

if __name__ == '__main__':
	headerList = ["Header1", "Header2", "Header3", "Header4"]	#array must be seperated by commas to be split into a list
	print (combinationGenerator([0,0], [1,1], [10,10], SpreadType="Addition",toFile=True,headerList=headerList))	#a list using these parameters is created and returned to comboList





### mumpce init script
import sys
import os

sys.path.append(os.path.dirname(__file__))

from Project import *
from measurement import tqfunc
#from response_surface import response_surface
#from solution import solution


import cantera as ct
import cantera.ck2cti as ck2cti
import canteraSimulate
import numpy as np
import canteraKineticsParametersParser 




    

def main():
    model_name = "ceO2"
    reactions_parameters_array = np.genfromtxt(model_name + "_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    listOfModifiableParameterIndices = canteraKineticsParametersParser.findModifiableParameterIndices(reactions_parameters_array)
    print("For this example, we get 9 parameters:")
    print(listOfModifiableParameterIndices)
    #[[0, 3], [0, 4], [0, 5], [1, 3], [1, 4], [1, 5], [1, 6], [1, 7], [1, 8]]   
    print("Now we make a silly example, just to demonstrate.")
    print("Here is before modification")
    print(reactions_parameters_array)
    modified_reactions_parameters_array= canteraKineticsParametersParser.modifyAllAdjustableReactionParametersArrayValues(reactions_parameters_array, [1,2,3,4,5,6,7,8,9])
    print("Here is after modification")
    print(modified_reactions_parameters_array)


if __name__ == '__main__':
    main()    


import numpy as np

class Solution(object):
    """A top level class for a constrained model with uncertainty
    
    :param solution_x: The solution vector
    :param covariance_x: The covariance matrix among the elements of the solution vector
    :param second_order_x: A structure describing the second order variation in the elements of the solution vector
    :type solution_x: ndarray,float
    :type covariance_x: ndarray,float
    :type second_order_x:
    
    """
    def __init__(self,
                solution_x,covariance_x=None,second_order_x=None,initial_x=None,initial_covariance=None):
        self.x     = solution_x #: The solution vector, :math:`x_{opt}`
        self.cov   = covariance_x #: The covariance matrix, :math:`\Sigma`
        self.alpha = np.linalg.cholesky(covariance_x) #: The lower triangular decomposition :math:`\alpha` where :math:`\Sigma = \alpha \alpha^{\text{T}}` 
        self.beta  = second_order_x
        
        self.x_i   = initial_x
        #"""The initial guess vector :math:`x_{init}` used to start the optimization, if not zero"""
        self.cov_i = initial_covariance
        #"""The initial covariance matrix :math:`\Sigma_{init}` used to start the optimization, if not zero"""
        self.alpha_i = None
        if initial_covariance is not None:
            self.alpha_i = np.linalg.cholesky(initial_covariance)
        
        
        return
    
    def update(self,new_x=None,new_cov=None):
        """Updates the solution and covariance in the Solution
        """
        if new_x is not None:
            self.x = new_x
        if new_cov is not None:
            self.cov = new_cov
            self.alpha = np.linalg.cholesky(new_cov)
        return


#These lines are to make sure that the module has access to canteraKineticsParametersParser
import os
import sys
file_dir = os.path.dirname(__file__)
sys.path.append(file_dir)

import cantera as ct
import numpy as np
import csv
from math import ceil
import canteraKineticsParametersParser 


def create_cti_and_SimulatePFRorTPRwithCantera(model_name, reactions_parameters_array, simulation_settings_module, cti_top_info_string = None, write_cti_to_file = False ):
    #The things that must be defined in advance are...
    # a) a model name.
    # b) The simulation settings are set in a python file which then becomes imported and an argument. This can then be changed by a script.
    # c) The reactions_parameters_array must be fed as an array or as a filename that points to a file that contains such an array with a header.
    # d) The cti_top_info_string. 
    cti_string, canteraPhases  = create_cti_and_cantera_phases(model_name=model_name, simulation_settings_module=simulation_settings_module, reactions_parameters_array=reactions_parameters_array, cti_top_info_string = cti_top_info_string, write_cti_to_file = write_cti_to_file)

    #Now import the simulation settings before running the simulation...
    concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
    simulatePFRorTPRwithCantera(model_name, canteraPhases['gas'], canteraPhases['surf'], simulation_settings_module)  
    return concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject
    
def create_cti_and_cantera_phases(model_name, reactions_parameters_array, simulation_settings_module, cti_top_info_string = None, write_cti_to_file = False ):
    #The things that must be defined in advance are...
    # a) a model name.
    # b) The simulation settings are set in a python file which then becomes imported and an argument. This can then be changed by a script.
    # c) The reactions_parameters_array must be fed as an array or as a filename that points to a file that contains such an array with a header.
    # d) The cti_top_info_string. 
    if type(reactions_parameters_array) == type("string"):
        reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    #In our example, the cti_top_info file is already made, and the functions below know where to find that file.
    
    #The below function makes a cti_file and then returns the filename.
    cti_string = canteraKineticsParametersParser.create_full_cti(model_name=model_name, reactions_parameters_array=reactions_parameters_array, cti_top_info_string = cti_top_info_string, write_cti_to_file = write_cti_to_file)
    
    canteraPhases = {}
    #NOTE: the reaction parameters are in the reaction objects. Currently, we cannot update the coverage dependence after cti file creation. That is why they must be created after the cti_file is made.
    # import the gas model and surface model.
    from distutils.version import LooseVersion, StrictVersion
    if LooseVersion(ct.__version__) < LooseVersion('2.5'):
        canteraPhases['gas'] = ct.Solution(source=cti_string, phaseid='gas')
        #canteraPhases['gas'].reactionsParametersArray = reactions_parameters_array
        # import the surface model
        canteraPhases['surf']  = ct.Interface(source=cti_string,phaseid='surf', phases=[canteraPhases['gas']]) #The word gas here is passing that same gas phase object in that was created above.
        #canteraPhases['surf'].reactionsParametersArray = reactions_parameters_array
    if LooseVersion(ct.__version__) >= LooseVersion('2.5'):
        canteraPhases['gas'] = ct.Solution(source=cti_string, name='gas')
        #canteraPhases['gas'].reactionsParametersArray = reactions_parameters_array
        # import the surface model
        canteraPhases['surf']  = ct.Interface(source=cti_string, name='surf', adjacent=[canteraPhases['gas']]) #The word gas here is passing that same gas phase object in that was created above.
        #canteraPhases['surf'].reactionsParametersArray = reactions_parameters_array
    #NOTE: we intentionally use "surf" rather than surface because in principle a person can make a model with more than 1 surface.
    #The choice of "surf" will make the variables easier to keep track of when it is time to make such a change.
    return cti_string, canteraPhases
    
#TODO: This can probably be generalized to run any simulation rather than only simulatePFRorTPRwithCantera.
def modify_reactions_and_SimulatePFRorTPRwithCantera(model_name, reactions_parameters_array, simulation_settings_module, canteraPhases, ArrheniusOnly = True, byProvidedReactionID = True):
    #The things that must be defined in advance are...
    # a) a model name.
    # b) The simulation settings are set in a python file which then becomes imported and an argument. This can then be changed by a script.
    # c) The reactions_parameters_array must be fed as an array or as a filename that points to a file that contains such an array with a header.
    # d) The cti_top_info_string. 
    # Note: for this function, what should be passed in is a modified version of eactions_parameters_array. Otherwise there is no point in using this function. 
    # If the original parameter array is going to be used, then simulatePFRorTPRwithCantera could be used directly.
    if type(reactions_parameters_array) == type("string"):
        reactions_parameters_array = np.genfromtxt(model_name + "_input_reactions_parameters.csv", delimiter=",", dtype="str", skip_header=1)
    #In our example, the cti_top_info file is already made, and the functions below know where to find that file.
    
    #The below function modifies the reactions in the gas phase.
    canteraKineticsParametersParser.modifyReactionsInOnePhase(canteraPhases['gas'], reactions_parameters_array, ArrheniusOnly = ArrheniusOnly, byProvidedReactionID = byProvidedReactionID)   
        #The below function modifies the reactions in the surface phase.
    canteraKineticsParametersParser.modifyReactionsInOnePhase(canteraPhases['surf'], reactions_parameters_array, ArrheniusOnly = ArrheniusOnly, byProvidedReactionID = byProvidedReactionID)
    
    #Now run the simulation...
    concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject = \
    simulatePFRorTPRwithCantera(model_name, canteraPhases['gas'], canteraPhases['surf'], simulation_settings_module)  
    return concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject
    
def simulatePFRorTPRwithCantera(model_name, canteraGasPhaseObject, canteraSurfacePhaseObject, simulation_settings_module):
    #simulation_settings_module_name must be a string if it is provided. The module itself should not be passed as an argument. This is intentional.
    canteraPhases ={}
    gas = canteraGasPhaseObject
    surf = canteraSurfacePhaseObject
    canteraPhases['gas'] = gas
    canteraPhases['surf'] = surf
       
    
    '''Now the code that handles *either* isothermal PFR or surface TPR. In future, will allow PFR TPR as well'''
    #We are going to model things as a flow reactor made of CSTRs, with no flow for the surface TPR case, following this example: https://cantera.org/examples/python/reactors/surf_pfr.py.html
    #it could probably also have been done as a flow reactor with no flow:  https://cantera.org/examples/python/surface_chemistry/catalytic_combustion.py.html
      

    
    '''START OF settings that users should change.'''
    flow_type = simulation_settings_module.flow_type
    T_gas_feed = simulation_settings_module.T_gas_feed
    T_surf = simulation_settings_module.T_surf
    velocity = simulation_settings_module.velocity
    reactor_cross_area = simulation_settings_module.reactor_cross_area
    cat_area_per_vol = simulation_settings_module.cat_area_per_vol
    porosity = simulation_settings_module.porosity
    length = simulation_settings_module.length
    P_gas = simulation_settings_module.P_gas
    gas_composition = simulation_settings_module.gas_composition
    t_step_size = simulation_settings_module.t_step_size
    t_final = simulation_settings_module.t_final
    NReactors = simulation_settings_module.NReactors
    print_frequency = simulation_settings_module.print_frequency
    heating_rate = simulation_settings_module.heating_rate
    surface_coverages =  simulation_settings_module.surface_coverages
    rtol = simulation_settings_module.rtol
    atol = simulation_settings_module.atol
    exportOutputs = simulation_settings_module.exportOutputs
    '''END OF settings that users should change.'''
    


    #Initiate concentrations output file and headers.
    if exportOutputs == True:
        concentrations_output_filename = model_name + "_output_concentrations.csv"
        outfile = open(concentrations_output_filename,'w')
        writer = csv.writer(outfile)
    concentrationsArrayHeaderList = ['Distance (m)', 'time(s)',  'T_gas (K)', 'T_surf (K)', 'P (atm)'] + \
                    gas.species_names + surf.species_names
    concentrationsArrayHeader = str(concentrationsArrayHeaderList)[1:-1] #The species names were imported when "surf" and "gas" objects were created.
    if exportOutputs == True:    
        writer.writerow(concentrationsArrayHeaderList)
    
    
    if flow_type == "Static":
        velocity = 0
        NReactors = 2 #In the surf_pfr example. For static, we only need 1 reactor, but the rlen formula below has a minimum value of 2. #FIXME
    
    num_t_steps = ceil(t_final / t_step_size) #rounds up.
    
    rlen = length/(NReactors-1) #This is each individual CSTR's length. #FIXME: Why is this not just Nreactors? Is it because of reservoirs?...
    #rvol is the reactor volume. We're modeling this as as a single CSTR with no flow.
    rvol = reactor_cross_area * rlen * porosity #Each individual CSTR gas volume reduced by porosity.
    # catalyst area in one reactor
    cat_area = cat_area_per_vol * rvol
    
    
    #Set the initial conditions for the gas and the surface.
    gas.TPX = T_gas_feed, P_gas, gas_composition
    surf.TP = T_surf, P_gas
    surf.X = surface_coverages
    
    mass_flow_rate = velocity * reactor_cross_area * gas.density  #linear velocity times area is volume/s, times kg/vol becomes kg/s. I think that for this example we neglect effects of surface adsorption regarding how much mass is in the gas phase?
    
    TDY = gas.TDY #Get/Set temperature [K] and density [kg/m^3 or kmol/m^3], and mass fractions. From: https://cantera.org/documentation/docs-2.4/sphinx/html/cython/thermo.html
    cov = surf.coverages #This is normalized coverages, built in from InerfacePhase class: https://cantera.github.io/docs/sphinx/html/cython/thermo.html#cantera.InterfacePhase
    #It would also be possible to type surface.site_density, giving [kmol/m^2] for surface phases. Also possible to use surf.set_unnormalized_coverages(self, cov) for cases when don't want to use normalized coverages.
    
    # create a new reactor
    gas.TDY = TDY #<-- If TDY was changing, and if original gas wanted to be kept, this could have been gas = copy.deepcopy(gas)?
    reactor = ct.IdealGasReactor(gas, energy='off')
    reactor.volume = rvol
    
    # create a reservoir to represent the reactor immediately upstream. Note
    # that the gas object is set already to the state of the upstream reactor
    upstream_of_CSTR = ct.Reservoir(gas, name='upstream') #A cantera reservoir never changes no matter what goes in/out: https://cantera.org/documentation/docs-2.4/sphinx/html/cython/zerodim.html#cantera.Reservoir
    
    # create a reservoir for the reactor to exhaust into. The composition of
    # this reservoir is irrelevant.
    downstream_of_CSTR = ct.Reservoir(gas, name='downstream') #A cantera reservoir never changes no matter what goes in/out: https://cantera.org/documentation/docs-2.4/sphinx/html/cython/zerodim.html#cantera.Reservoir
    
    #Note: these are reservoirs for individual CSTRs. It's used in a clever way in this example, as will be seen later.
    #I would prefer to call these reservoirs "feed" and "exhaust", and would consider to fill the downstream/exhaust with inert to make it clear that it is different. Or maybe make deep copies of gas called "feed_gas" and "exhaust_gas".
    
    # Add the reacting surface to the reactor. The area is set to the desired
    # catalyst area in the reactor.
    rsurf = ct.ReactorSurface(surf, reactor, A=cat_area)  #Here is where the catalyst site density gets used.  Note that cat_area is actually in meters now, even though site_density is defined as mol/cm^2 in the cti file.
    
    # The mass flow rate into the reactor will be fixed by using a
    # MassFlowController object.
    f_mfr = ct.MassFlowController(upstream_of_CSTR, reactor, mdot=mass_flow_rate) #mass flow controller makes flow that goes from the first argument to the second argument with mdot providing "The mass flow rate [kg/s] through this device at time t [s]."s
    
    # We need an outlet to the downstream reservoir. This will determine the
    # pressure in the reactor. The value of K will only affect the transient
    # pressure difference.
    e_mfr = ct.PressureController(reactor, downstream_of_CSTR, master=f_mfr, K=1e-5)  #This makes flow that goes from first argument to second argument with the mass flow controller "f_mfr" controlling the pressure here. K has units of kg/s/Pa times the pressure difference. this "v" that comes out is in same units as the mass flow rate, it's in kg/s.   https://cantera.org/documentation/docs-2.4/sphinx/html/cython/zerodim.html#flowdevice e_mfr is the exhaust mass flow rate.
    
    sim = ct.ReactorNet([reactor]) #This is normally a list of reactors that are coupled.
    sim.max_err_test_fails = 12 #Even after looking at docs, I'm not sure what this does. I think this is how many relative and absolute tolerance failures there can be in a single time step before the resolution of the ODE integrator does that step again with finer resolution? https://cantera.org/documentation/docs-2.4/sphinx/html/cython/zerodim.html#reactor-networks
    
    #Static versus PFR flag
    
    
    
    # set relative and absolute tolerances on the simulation
    sim.rtol = rtol
    sim.atol = atol
    gas_rates = [] #NOTE: These are just the rates from the surface phase. These are *not* including any homogeneous rates.
    surface_rates = [] #NOTE: These are just the rates from the surface phase. These are *not* including any homogeneous rates.
    sim_times = [] #NOTE: This is less useful if one uses advance_to_steady_state
    sim_dist = [] #This is the distance in the reactor. If one is using flow and advance_to_steady_state, then this is representative of the kinetics.
    concentrationsArray = []
    #Print some things out for before the simulation, these are basically headers.
    if print_frequency != None:
        print(concentrationsArrayHeader)
    
    if flow_type == "PFR":
        for n in range(NReactors): #iterate across the CSTRs.
            # Set the state of the reservoir to match that of the previous reactor
            gas.TDY = reactor.thermo.TDY #<-- setting gas to *most recent* state of the reactor.
            upstream_of_CSTR.syncState() #<-- this is an interesting trick. Once one CSTR has been simulated, we set the current values (the one from the just simulated CSTR) to be upstream value. This works because we're doing implicit ODE down a reactor length, and probably only gives the right answer for steady state kinetics. It also only works because we don't care what's happening downstream. We have a fixed K for the e_mfr, and also f_mfr.
            sim.reinitialize()
            sim.advance_to_steady_state() #<-- we advance the current CSTR to steady state before going to the next one.  For a tpr, we'd probably use advance(self, double t) instead. The syntax would be something like sim.advance(10.0) for 10 seconds. https://cantera.github.io/docs/sphinx/html/cython/examples/reactors_reactor1.html#py-example-reactor1-py
        
            dist = n * rlen # distance in m <-- this could have been defined above, and I would have... but it's probably good to have it down here to make it clear that only 'n" is being used rather than this value, to simulate the reactor.
            sim_dist.append(dist)
            sim_times.append(sim.time)
            gas_rates.append(surf.get_net_production_rates('gas'))
            surface_rates.append(surf.get_net_production_rates('surf'))
            # write the gas mole fractions and surface coverages vs. distance
            rowListOfStrings = [dist, sim.time, gas.T, surf.T, reactor.thermo.P/ct.one_atm] + \
                        list(gas.concentrations) + list(surf.coverages)
            if exportOutputs == True:
                writer.writerow(rowListOfStrings)
                concentrationsArray.append(np.array(rowListOfStrings))
            if print_frequency != None: 
                if not n % print_frequency: #This only prints every specified number of steps.
                    try:
                        with np.printoptions(precision=3):
                            print(np.array(rowListOfStrings))
                    except:
                            print(np.array(rowListOfStrings))
    
    if flow_type == "Static":
        T_surf_0 = T_surf
        #sim.set_max_time_step(t_step_size) #Cantera's sim.step normally advances in a variable way.
        for i_step in range(num_t_steps): #iterate across the CSTRs.
            time = i_step*t_step_size
            T_surf = T_surf_0 + time*heating_rate
            surf.TP = T_surf, P_gas
            if simulation_settings_module.piecewise_coverage_dependence == True:
                modified_reactions_parameters_array = canteraKineticsParametersParser.calculatePiecewiseCoverageDependentModifiedParametersArray(simulation_settings_module, surf.species_names, surf.coverages) #This feature requires the piecewise coverage dependence settings AND the reactions_parameters_array to already be inside the surf object **in advance**
                canteraKineticsParametersParser.modifyReactionsInOnePhase(surf, modified_reactions_parameters_array, ArrheniusOnly=True) #TODO: Right now this is constrainted to Arrhenius only because cantera does not yet allow modifyReactionsInOnePhase to do more (but it's planned to change in developers roadmap)                
            surf.advance_coverages(t_step_size)  #sim.advance(time) would not work. Changing T with time is not officially supported but happens to work with surf.advance_coverages. Supported way to change temperature during simulation for arbitrary reactors is to use custom integrator: https://cantera.org/examples/python/reactors/custom.py.html
            dist = 0.0
            sim_dist.append(dist)
            sim_times.append(time)
            gas_rates.append(surf.get_net_production_rates('gas'))
            surface_rates.append(surf.get_net_production_rates('surf'))        
            rowListOfStrings = [dist, time, gas.T, surf.T, reactor.thermo.P/ct.one_atm] + \
                    list(gas.X) + list(surf.coverages)
            if exportOutputs == True:
                writer.writerow(rowListOfStrings)
                concentrationsArray.append(np.array(rowListOfStrings))
            if print_frequency != None:
                if not i_step % print_frequency: #This only prints every specified number of steps.
                    try:
                        with np.printoptions(precision=3):
                            print(np.array(rowListOfStrings))
                    except:
                            print(np.array(rowListOfStrings))

           
    #Need to get things into a stackable state. Figured out from numpy shape that I needed to do at_least2D and transpose.
    sim_dist =  np.atleast_2d(sim_dist).transpose()
    sim_times =  np.atleast_2d(sim_times).transpose()
    gas_rates =  np.array(gas_rates)
    surface_rates =  np.array(surface_rates)    
    gasRatesArray = np.hstack((sim_dist,sim_times, gas_rates))
    surfaceRatesArray = np.hstack((sim_dist,sim_times, surface_rates))
    rates_all_array = np.hstack((sim_dist,sim_times, gas_rates, surface_rates))
    concentrationsArray = np.array(concentrationsArray)
    concentrationsArrayHeader = concentrationsArrayHeader
    gasRatesArrayHeader = 'dist(m), time(s),'+str(gas.species_names).replace("'","")[1:-1]
    surfaceRatesArrayHeader = 'dist(m),time(s),'+str(surf.species_names).replace("'","")[1:-1]
    rates_all_array_header = 'dist(m),time(s),'+str(gas.species_names).replace("'","")[1:-1]+"," + str(surf.species_names).replace("'","")[1:-1]
    canteraSimulationsObject = sim
    cantera_phase_rates = {"gas":gasRatesArray, "surf":surfaceRatesArray}
    cantera_phase_rates_headers = {"gas":gasRatesArrayHeader, "surf":surfaceRatesArrayHeader}
    
    if exportOutputs == True:
        np.savetxt(model_name + "_output_rates_all.csv", rates_all_array, delimiter=",", comments = '', header = rates_all_array_header)
        np.savetxt(model_name + "_output_rates_gas.csv", gasRatesArray, delimiter=",", comments = '', header = gasRatesArrayHeader )
        np.savetxt(model_name + "_output_rates_surf.csv", surfaceRatesArray, delimiter=",", comments = '', header = surfaceRatesArrayHeader)     
    if exportOutputs == True:    
        outfile.close()
    return concentrationsArray, concentrationsArrayHeader, rates_all_array, rates_all_array_header, cantera_phase_rates, canteraPhases, cantera_phase_rates_headers, canteraSimulationsObject


import numpy as np
import pickle
from response_surface import ResponseSurface

def idfunc(*arg,**kwargs):
    if len(arg) == 1:
        return arg[0]
    return arg

#print('loading')

try:
    import tqdm
    tqfunc = tqdm.tqdm_notebook
    #print('tqdm found')
except ImportError:
    #is not available
    tqfunc = idfunc

class Measurement(object):
    """A top level class for a measurement object
    
    This class is intended to contain metadata for the measurement it defines as well as a simulation model and a response 
    surface for optimization.
    
    :param name: The name of the measurement
    :param comment: An arbitrary text string describing the measurement
    :param model: The simulation model
    :param value: The measured value from the experiment
    :param uncertainty: The uncertainty in the measured value
    :param active_parameters: The list of active parameters for this simulation. Not normally defined at object creation
    :param parameter_uncertainties: The list of uncertainties in the active parameters. Not normally defined at object creation
    :param response: The response surface for this simulation. Not normally defined at object creation
    :param response_type: Whether the response is to be linear ('linear', :math:`y = a^{\\text{T}}x + z`) or logarithmic ('log', :math:`\ln y = a^{\\text{T}}x + z`)
    :param response_perturbation: The perturbation in the parameters for response surface generation
    :param model_value: The computed value from the simulation model for this measurement
    :param sensitivity_list: The list of sensitivities of the model value to each parameter
    :type name: str
    :type comment: str
    :type model: model object
    :type value: float
    :type uncertainty: float
    :type active_parameters: int list
    :type parameter_uncertainties: float list
    :type response: response surface object
    :type response_type: str
    :type response_perturbation: float
    :type model_value: float
    :type sensitivity_list: float list
    
    """
    def __init__(self,
                 name=None,
                 model=None,
                 value=None,
                 uncertainty=None,
                 active_parameters=None,
                 parameter_uncertainties=None,
                 response=None,
                 response_type='linear',
                 response_perturbation=0.3,
                 model_value=None,
                 sensitivity_list=None,
                 comment=None
                ):
        
        self.tqfunc = tqfunc
        
        #Define the name
        self.name = name
        self.comment = comment
        
        #Define the computer model that will be used to simulate this measurement
        self.model = model
        
        
        #Define the experimental value
        self.value = value
        self.uncertainty = uncertainty
        
        #Define the status of the experiment (used by Project for reporting purposes)
        self._status = 'Active'
        
        #Define the response surface and active parameters for this measurement
        self.active_parameters = active_parameters
        self.response = response
        self.parameter_uncertainties = parameter_uncertainties
        
        #Define the perturbations for sensitivity analysis
        self.response_perturbation = response_perturbation
        self.response_sensitivity = 1.0e-3
        self.response_type = response_type
        
        #Take the sensitivity list from the initialization
        self.model_value = model_value
        self.model_uncertainty = None
        self.sensitivity_list=sensitivity_list
        
        #Create attributes for optimized values and uncertainties
        self.optimized_value = None
        self.optimized_uncertainty = None
        self.consistency = None
        self.weighted_consistency = None
        self.entropy = None
        
        return
    
    def __str__(self):
        """Returns Name (Status): str(self.model)
        """
        modelstr = str(self.model)
        
        meas_str = "{} ({}): {}".format(self.name,self._status,modelstr,)
        
        if self.comment:
            meas_str = meas_str + "\n     {}".format(self.comment)
        
        return meas_str
        
    def set_value(self,value,uncertainty):
        """Sets the experimental value and uncertainty for this measurement
        
        :param value: The experiment's measured value
        :param uncertainty: The measurement uncertainty in the value
        :type value: float
        :type uncertainty: float
        
        """
        self.value = value
        self.uncertainty = uncertainty
        return
    
    def make_response(self): #(self,zero_term,perterbations,sensitivities):
        """Generates a sensitivity_analysis_based response surface for this measurement
        """
        #zero_term = self.evaluate
        self.model.reset_model()
        logfile_name = self.name + '_resp_log.out'
        response_logfile = open(logfile_name,'w')
        
        sensitivity_args = (self.response_sensitivity,
                           self.active_parameters,
                           response_logfile)
        sensitivity_kw = dict(tq=True)
        
        zero_term, sens_zero = self.model.sensitivity(*sensitivity_args)
        
        if self.response_type == 'log':
            zero_term = np.log(zero_term)
        
        #print zero_term
        #print sens_zero
        
        number_params = len(self.active_parameters)
        
        #Calculate the multipliers that will be used for the SAB sensitivity calculations
        multipliers = self.parameter_uncertainties ** self.response_perturbation
        #print multipliers
        
        perturbations = np.zeros((number_params,2))
        sens_positive = np.zeros((number_params,number_params))
        sens_negative = np.zeros_like(sens_positive)
        
        #Changed this so that tqdm will be used if it is available, but otherwise not
        #for (parameter_number,parameter) in tqdm.tqdm(enumerate(self.active_parameters)):
        for (parameter_number,parameter) in enumerate(self.tqfunc(self.active_parameters)):
            self.model.reset_model()
            base_value = self.model.get_parameter(parameter)
            param_name = self.model.model_parameter_info[parameter]['parameter_name']
            
            #print 'Parameter = ', parameter
            
            
            positive_perturbation = multipliers[parameter_number]
            negative_perturbation = 1/positive_perturbation
            
            #print positive_perturbation
            #print negative_perturbation
            
            #Positive perturbation
            response_logfile.write('\nParameter number = {: 4d} {:30s}\n'.format(parameter,param_name))
            response_logfile.write('Positive perturbation = {: 10.5e}\n'.format(positive_perturbation))
            self.model.perturb_parameter(parameter,positive_perturbation*base_value)
            value_pos, sens_pos = self.model.sensitivity(*sensitivity_args,**sensitivity_kw)
            
            #Negative perturbation
            response_logfile.write('Negative perturbation = {: 10.5e}\n'.format(negative_perturbation))
            self.model.perturb_parameter(parameter,negative_perturbation*base_value)
            value_neg, sens_neg = self.model.sensitivity(*sensitivity_args,**sensitivity_kw)
            
            perturbations[parameter_number,:] = [value_pos, value_neg]
                
            sens_positive[parameter_number,:] = sens_pos
            sens_negative[parameter_number,:] = sens_neg
            
            #print "Positive sensitivity:", sens_pos
            #print "Negative sensitivity:", sens_neg
            #print ''
        
        self.model.reset_model()
        #First order terms of response surface
        if self.response_type == 'log':
            perturbations = np.log(perturbations)
        a_terms = (perturbations[:,0] - perturbations[:,1]) / (2 * self.response_perturbation)
        #Second order terms of response surface
        b_terms_first = (sens_positive - sens_negative) * np.log(self.parameter_uncertainties) / (4 * self.response_perturbation)
        
        if self.response_type == 'linear':
            #a_terms = a_terms * zero_term
            b_terms_first = b_terms_first * zero_term
        
        b_terms = (b_terms_first + b_terms_first.T)/2# - np.diag(np.diag(b_terms_first))
        d_terms = (b_terms_first - b_terms_first.T)/2
        
        
        self.response = ResponseSurface(zero_term=zero_term,
                                         a_terms=a_terms,
                                         b_terms=b_terms,
                                         d_terms=d_terms,
                                         active_parameters=self.active_parameters)
        return
    
    def evaluate(self):
        """Evaluates the model once and sets self.model_value to the returned value.
        
        :returns: model_value
        """
        self.model_value = self.model.evaluate()
        if self.response_type == 'log':
            self.model_value = np.log(self.model_value)
        return self.model_value
    
    def evaluate_sensitivity(self,perturbation=0.05):
        """Conducts a sensitivity analysis on the model and storee the nominal value in self.model_value and the sensitivity in self.sensitivity_list
        
        :param perturbation: The amount to perturb each parameter when conducting the sensitivity analysis
        :type perturbation: float
        """
        all_parameters = np.arange(self.model.number_parameters,dtype=int)
        logfile_name = self.name + '_sen.out'
        with open(logfile_name,'w') as logfile:
                self.model_value,self.sensitivity_list = self.model.sensitivity(perturbation=perturbation,
                                                                                parameter_list=all_parameters,
                                                                                logfile=logfile
                                                                               )
        if self.response_type == 'log':
            self.model_value = np.log(self.model_value)
        return
    
    def print_sorted_sensitivity(self,sensitivity=None,max_number=None):
        """Sorts the parameters by sensitivity coefficient and prints them.
        
        :param sensitivity: The list of sensitivity coefficients. If None, will use self.sensitivity_list
        :param max_number: If not None, will only print the top max_number parameters
        :type sensitivity: array_like or None
        :type max_number: int or None
        """
        
        if sensitivity is None:
            sensitivity = self.sensitivity_list
        
        sorted_param_nums = np.argsort(abs(sensitivity))
        
        if max_number is None:
            max_number = self.model.number_parameters
        
        print_params = sorted_param_nums[-1:-1*max_number:-1]
        
        for print_param in print_params:
            param_info = self.model.model_parameter_info[print_param]
            param_name = param_info['parameter_name']
            print('{: 4d} {: 10.4e}  {}'.format(print_param,
                                                sensitivity[print_param],
                                                param_name
                                               )
                )
            
           
        return
    
    def evaluate_response(self,x):
        """Evaluates the response surface for this measurement.
        
        :param x: The set of parameter values at which the surface must be evaluated
        :type x: 1d array
        :returns: response
        :rtype: float
        """
        response = self.response.evaluate(x)
        return response
    
    def sensitivity_response(self,x):
        """Evaluates the response surface and the response surface gradient for this measurement.
        
        :param x: The set of parameter values at which the surface must be evaluated
        :type x: 1d array
        :returns: response,response_gradient
        :rtype: tuple
        """
        response,response_grad = self.response.sensitivity(x)
        return response,response_grad
    
    def evaluate_uncertainty(self,x,cov):
        """Evaluates the response surface for this measurement and compute its uncertainty
        
        :param x: The set of parameter values at which the surface must be evaluated
        :param cov: The covariance matrix among the model parameters
        :type x: 1d array
        :type cov: 2d array
        :returns: response,response_uncertainty
        :rtype: tuple
        """
        response,response_uncertainty = self.response.evaluate(x,cov)
        return response,response_uncertainty
    
    def save(self):
        """Saves a pickled representation of the measurement
        """
        filename = self.name + '.save'
        
        with open(filename,'wb') as f:
            pickle.dump(self,f)
        return
    
    def prepare_for_save(self):
        self.model.prepare_for_save()
        return
    
    def _save(self):
        """Saves the model value, sensitivity list, and response surface to disk. By default, they are saved to the file 
        'self.name'.npz
        """
        
        #The response surface may not exist, so create some dummy variables to ensure that the measurement can be saved
        localz = None
        locala = None
        localb = None
        locald = None
        
        print('Checking for response surface ...')
        if self.response is not None:
            print('Response surface exists, saving response surface ...')
            localz = self.response.z
            locala = self.response.a
            localb = self.response.b
            locald = self.response.d
        
        outputfilename = self.name + '.npz'
        print ('Saving to output file: {}'.format(outputfilename))
        with open(outputfilename, 'w+') as outputfile:
            outputfile.seek(0)
            np.savez(outputfile,
                     model_value = self.model_value,
                     sensitivity_list = self.sensitivity_list,
                     z = localz,
                     a = locala,
                     b = localb,
                     d = locald,
                     active_parameters = self.active_parameters,
                     active_parameter_uncertainties = self.parameter_uncertainties,
                     value=self.value,
                     uncertainty=self.uncertainty
                    )
        return
    
    def load(self):
        """Loads the model value, sensitivity list, and response surface from disk. By default, they are loaded from the 
        file 'self.name'.npz
        """
        inputfilename = self.name + '.npz'
        print ('Loading from output file: {}'.format(inputfilename))
        with open(inputfilename, 'r') as inputfile:
            response_data = np.load(inputfile)
            
            #Check to see if the file has response surface data in it
            localz = response_data['z']
            print('Checking response surface...')
            #Do not load the response surface if there is no response surface data
            if localz is not None:
                print('Response surface exists, loading response surface ...')
                self.response = ResponseSurface(zero_term=response_data['z'],
                                                 a_terms=response_data['a'],
                                                 b_terms=response_data['b'],
                                                 d_terms=response_data['d'],
                                                 active_parameters=response_data['active_parameters']
                                                )
            self.model_value = response_data['model_value']
            if self.model_value: print('Model value loaded')
                
            self.sensitivity_list = response_data['sensitivity_list']
            if self.sensitivity_list.any(): print('Sensitivity vector loaded')
                
            self.value = response_data['value']
            if self.value: print('Experimental value loaded')
                
            self.uncertainty = response_data['uncertainty']
            if self.uncertainty: print('Experimental uncertainty loaded')
                
            self.active_parameters = response_data['active_parameters']
            if self.active_parameters.any(): print('Active parameters loaded')
                
            self.parameter_uncertainties = response_data['active_parameter_uncertainties']
            if self.parameter_uncertainties.any(): print('Parameter uncertainties loaded')
                
        return
    
    def print_model_values(self):
        parameter_info = self.model.model_parameter_info
        
        headname = 'Parameter name'
        headv = 'Value'
        headf = 'Uncert'
        headx = 'FactVal'
        heads = 'FactUnc'
        headfs = 'NewVal'
        headsig = 'Uncert'
        
        head_args = (headname,headv,headf)
        
        head_format = '{:40s}   {:8s} {:7s}'
        
        #print(head_format.format(*head_args))
        
        header = head_format.format(*head_args)
        
        carriage_return = '\n'
        output = ''
        
        output = carriage_return.join((output,header))
        
        for active_num,param in enumerate(self.active_parameters):
            
            param_name = parameter_info[param]['parameter_name']
            
            value = parameter_info[param]['parameter_value']#self.model.get_parameter(param)
            this_unc = self.parameter_uncertainties[active_num]
            
            print_args = (param_name[:40],value,this_unc)
            
            print_string = '{:40s} : {: 8.2g} {: 7.2f}'
            #if abs(value) > 5000:
            #    print_string = '{:30s} : {: 7.2e} {: 7.2f}'
            
            #print(print_string.format(*print_args))
            line = print_string.format(*print_args)
            output = carriage_return.join((output,line))
        
        return output
    
    def get_active_names(self):
        names = []
        for param in self.active_parameters:
            param_name = self.model.model_parameter_info[param]['parameter_name']
            names += [param_name]
        return names
    
    def interpret_model(self,x,cov):
        parameter_info = self.model.model_parameter_info
        
        headname = 'Parameter name'
        headv = 'Value'
        headf = 'Uncert'
        headx = 'FactVal'
        heads = 'FactUnc'
        headfs = 'NewVal'
        headsig = 'Uncert'
        
        head_args = (headname,headv,headf,headx,heads,headfs,headsig)
        
        head_format = '{:40s}   {:8s} {:7s} {:7s} {:7s} {:8s} {:7s}'
        
        #print(head_format.format(*head_args))
        
        carriage_return = '\n'
        output = ''
        
        header = head_format.format(*head_args)
        output = carriage_return.join((output,header))
        
        for active_num,param in enumerate(self.active_parameters):
            
            param_name = parameter_info[param]['parameter_name']
            
            value = parameter_info[param]['parameter_value']#self.model.get_parameter(param)
            
            this_x = x[active_num]
            this_std = 2*np.sqrt(cov[active_num,active_num])
            this_unc = self.parameter_uncertainties[active_num]
            
            multiplier = this_unc ** this_x
            new_value = value*multiplier
            new_uncertainty = this_unc ** (this_std)
            
            print_args = (param_name[:40],value,this_unc,this_x,this_std,new_value,new_uncertainty)
            
            print_string = '{:40s} : {: 8.2g} {: 7.2f} {: 7.2f} {: 7.2f} {: 8.2g} {: 7.2f}'
            #if abs(value) > 5000:
            #    print_string = '{:30s} : {: 7.2e} {: 7.2f} {: 7.2f} {: 7.2f} {: 7.2e} {: 7.2f}'
            line = print_string.format(*print_args)
            output = carriage_return.join((output,line))
            #print(print_string.format(*print_args))
        
        return output
    
    def modify_model(self,x):
        for active_num,param in enumerate(self.active_parameters):
            #value = self.model.get_parameter(param)
            this_x = x[active_num]
            this_unc = self.parameter_uncertainties[active_num]
            
            multiplier = this_unc ** this_x
            new_value = multiplier#*value
            
            self.model.perturb_parameter(param,new_value)
        return
    
    def get_model_values(self):
        values = []
        uncertainties = []
        parameter_info = self.model.model_parameter_info
        for active_num,param in enumerate(self.active_parameters):
            value = parameter_info[param]['parameter_value']#self.model.get_parameter(param)
            this_unc = self.parameter_uncertainties[active_num]
            values += [value]
            uncertainties += [this_unc]
        values = np.array(values)
        uncertainties = np.array(uncertainties)
        return values,uncertainties
    
    def get_opt_values(self,x,cov):
        new_values = []
        new_uncertainties = []
        parameter_info = self.model.model_parameter_info
        for active_num,param in enumerate(self.active_parameters):
            value = parameter_info[param]['parameter_value']#self.model.get_parameter(param)
            this_x = x[active_num]
            this_std = 2*np.sqrt(cov[active_num,active_num])
            this_unc = self.parameter_uncertainties[active_num]

            multiplier = this_unc ** this_x
            new_value = value*multiplier
            new_uncertainty = this_unc ** (this_std)

            new_values += [new_value]
            new_uncertainties += [new_uncertainty]
        new_values = np.array(new_values)
        new_uncertainties = np.array(new_uncertainties)

        return new_values,new_uncertainties
            


import copy
import math
import numpy as np

class ResponseSurface(object):
    """A top level class describing a polynomial response surface.
    
    This class contains the methods for a response surface object. It largely parallels :func:`model` in that it has an :func:`evaluate` method which will return a value and a :func:`sensitivity` method which will return a value and a vector of sensitivities.
    
    The response surface is assumed here to be a second order polynomial :math:`y = z + a^{\\text{T}}x + x^{\\text{T}}bx`
    
    :param zero_term: The zero-order term of the response surface, :math:`z`
    :param a_terms: The first order terms of the response surface, :math:`a`
    :param b_terms: The second order terms of the response surface, :math:`b`
    :param c_terms: The third order terms of the response surface (not implemented)
    :param d_terms: The estimated error in the second order terms of the response surface
    :param active_parameters: If the list of active paremeters differs from measurement to measurement, it would be specified here.
    
    :type zero_term: float
    :type a_terms: ndarray(float), len(active_parameters)
    :type b_terms: ndarray(float), len(active_parameters)xlen(active_parameters)
    :type c_terms: ndarray(float)
    :type d_terms: ndarray(float), len(active_parameters)xlen(active_parameters)
    :type active_parameters: ndarray(int)
    
    
    :todo: implement active_parameters, implement c_terms
    
    
    """
    def __init__(self,
                 zero_term=None,
                 a_terms=None,b_terms=None,c_terms=None,
                 d_terms=None,
                 active_parameters=None):
        
        self.z = zero_term #: :math:`z`        
        self.a = a_terms #: :math:`a` 
        self.b = b_terms #: :math:`b` 
        
        self.c = c_terms
        self.d = d_terms
        self.active_parameters = active_parameters
        return
    
    def evaluate(self,x,cov_x=None):#x_full,exp_value,weight):
        """ Evaluates the response surface
        
        Computes the response value :math:`y = z + a^{\\text{T}}x + x^{\\text{T}}bx`
        If cov_x is supplied, returns the uncertainty :math:`\\sigma^2 = a^{\\text{T}}\\Sigma a + 2\\text{tr}((b\\Sigma)^2)``
        
        :param x: The parameter vector at which the response value is to be calculated
        :param cov_x: The covariance matrix among the parameters
        :type x: ndarray(float), len(active_parameters)
        :type cov_x: ndarray(float), len(active_parameters)xlen(active_parameters)
        :returns: response_value: :math:`y = z + a^{\\text{T}}x + x^{\\text{T}}bx` and  computed_unc: :math:`\\sigma^2 = a^{\\text{T}}\\Sigma a + 2\\text{tr}((b\\Sigma)^2)` if cov_x is supplied
        :rtype: tuple
        
        """
        #Select only the active parameters from the full parameter list
        #x = x[active_parameters]
        
        
        #Evaluate the response surface
        #Zero and first order terms
        response_value  = copy.deepcopy(self.z)
        response_value += np.dot(self.a,x)
        
        if cov_x is not None:
            a_times_cov = np.dot(self.a,cov_x)
            variance = np.dot(self.a,a_times_cov.T)
        
        #Second order terms (might not exist)
        if self.b is not None:
            b_times_x = np.dot(self.b,x)
            response_value += np.dot(b_times_x.T,x)
            if cov_x is not None:
                b_times_cov = np.dot(self.b,cov_x)
                variance += 2*np.trace(np.dot(b_times_cov,b_times_cov))
              
        #Third order terms not implemented
        
        if cov_x is not None:
            computed_unc = math.sqrt(variance)
            return response_value,computed_unc
        
        
        return response_value
    
    def sensitivity(self,x):
        """ Evaluates the response surface and response surface gradient
        
        Computes the response value :math:`y = z + a^{\\text{T}}x + x^{\\text{T}}bx` and the response surface gradient :math:`\\frac{dy}{dx} = a + 2bx`
        
        :param x: The parameter vector at which the response value is to be calculated
        :type x: ndarray(float), len(active_parameters)
        :returns: response_value: :math:`y = z + a^{\\text{T}}x + x^{\\text{T}}bx` and response_gradient: :math:`\\frac{dy}{dx} = a + 2bx`
        :rtype: tuple
        
        """
        #Select only the active parameters from the full parameter list
        #x = x[active_parameters]
        
        #Evaluate the response surface
        #Zero and first order terms
        response_value  = copy.deepcopy(self.z)
        response_value += np.dot(self.a,x)
        
        response_grad = copy.deepcopy(self.a)
        
        #Second order terms (might not exist)
        if not(self.b is None):
            b_times_x = np.dot(self.b,x)
            response_value += np.dot(b_times_x.T,x)
            response_grad += 2*b_times_x
              
        #Third order terms not implemented

        return response_value,response_grad


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 28 14:44:31 2019

@author: Yurik
"""

import xmltodict
import numpy as np
import cantera as ct 
import copy
#This module is written for reactions_parameters_arrays in the following format:
#headerString = "reactionID,canteraReactionType,reactionEquation,A,b,E,a,m,e,coverageDependenceSpecies, is_sticking_coefficient"
#The csv files are comma separated, and have 1 header string as noted above.


def stackListsAndArrays(listOfItemsToStack):
    stackableItemsList = copy.deepcopy(listOfItemsToStack)
    for itemIndex in range(len(listOfItemsToStack)):
        #if it is a list, we turn it into a 2D array of the appropriate geometry.
        if str(type(listOfItemsToStack[itemIndex]))=="<class 'list'>":
            stackableItemsList[itemIndex] = np.atleast_2d(stackableItemsList[itemIndex]).transpose()
        else:
            pass
    return np.hstack(tuple(stackableItemsList)) #convert it to array and stack it. hstack requires a tuple as an argument.


def descendingLinearEWithPiecewiseOffsetCheckOneReactionAllReactions(reactions_parameters_array, piecewise_coverage_intervals_all_reactions, E_offsets_array_all_reactions, verbose = False):
    #Note that E_0 and g_slope are contained inside reactions_parameters_array
    #piecewise_coverage_intervals_all_reactions can either be 1D (defined the same for all reactions) or can be 2D (different intervals for each reaction)
    #piecewise_coverage_intervals_all_reactions and E_offsets_array_all_reactions can be staggered if they are 2D, but must have one row for each reaction.
    #NOTE: this calculates a single E_offsets_array_all_reactions, but actually there is going to be one for each species. The complication is that during simulations, there are many different coverage combinations possible
    #It would probably be better to sample a grid of possible species coverages that might be encountered, but for now this function will work for only one species.
    passedArray = np.zeros(len(reactions_parameters_array), dtype="bool") #This is an array that starts as all False, and we fill with true for each reaction that passes.
    for reactionIndex in range(len(reactions_parameters_array)):
        individualreactions_parameters_array = reactions_parameters_array[reactionIndex]  #TODO: it might be better to pass the individual reaction parameters array to the checking function, which would mean changing the checking function.
        if len(np.shape(reactions_parameters_array))==1:
            piecewise_coverage_intervals = piecewise_coverage_intervals_all_reactions
        else:
            piecewise_coverage_intervals = piecewise_coverage_intervals_all_reactions[reactionIndex]
        E_offsets_array = E_offsets_array_all_reactions[reactionIndex]
        
        #Populate the Arrhenius type parameters related to E and its slope...
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        individualReactionTypeReceived = str(individualreactions_parameters_array[1])
        reactionEquation = str(individualreactions_parameters_array[2])
#        A = str(individualreactions_parameters_array[3])
#        b = str(individualreactions_parameters_array[4])
        E = str(individualreactions_parameters_array[5])
#        a = str(individualreactions_parameters_array[6])
#        m = str(individualreactions_parameters_array[7])
        e = str(individualreactions_parameters_array[8])
        coverageDependenceSpecies = str(individualreactions_parameters_array[9])

        #Get the numbers that we need out.
        E_0 = float(E)
        
        if np.isnan(float(e)): #Need to make e into 0.0 if it's an nan. Otherwise the descending check will fail.
            e = 0.0
        g_slope = -1.0*float(e) #Note that in cantera e is the negative of g_slope! https://cantera.org/science/reactions.html
        print(reactionID, reactionEquation, e, g_slope)
        if individualReactionTypeReceived != "surface_reaction":
            passedArray[reactionIndex] = True  #if it's not a surface reaction, we are not supposed to check it.
        if individualReactionTypeReceived == "surface_reaction":
            passedArray[reactionIndex] = descendingLinearEWithPiecewiseOffsetCheckOneReaction(E_0, g_slope,piecewise_coverage_intervals, E_offsets_array)
            if verbose == True:
                if passedArray[reactionIndex] == False:
                    print("The reaction with ID of" + reactionID + "and equation of" + reactionEquation + "did not pass descendingLinearEWithPiecewiseOffsetCheckOneReaction" )
    if sum(passedArray )< len(passedArray): #For a boolean array, if the sum is less than the length, that means at least one thing remained false.
        return False
    elif sum(passedArray) == len(passedArray): #Else the sum must be equal and we return true.
        return True
    
def descendingLinearEWithPiecewiseOffsetCheckOneReaction(E_0, g_slope,piecewise_coverage_intervals, E_offsets_array):
    E_array =E_0 + g_slope*piecewise_coverage_intervals+E_offsets_array
    delta_E_array = np.diff(E_array)
    if np.any(delta_E_array > 0):
        return False
    else:
        return True

def makeCanteraReactionObjectsListFromFile(FileName):
    canteraReactionObjectsList = ct.Reaction.listFromFile(FileName)
    return canteraReactionObjectsList


def extractReactionParametersFromFile(InputFileName, OutputFilename = ""): #The input fileName must point to a cti file or an xml file. An optional argument of OutputFilename will export the reactionParamtersArray
    canteraReactionObjectsList = makeCanteraReactionObjectsListFromFile(InputFileName)
    reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList, is_sticking_coefficientList = getReactionParametersFromCanteraReactionObjectsList(canteraReactionObjectsList)
    outputAsNumpyArray = stackListsAndArrays([reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList, is_sticking_coefficientList])
    if OutputFilename != "":
        headerString = "reactionID,canteraReactionType,reactionEquation,A,b,E,a,m,e,coverageDependenceSpecies, is_sticking_coefficient"
        np.savetxt(OutputFilename,outputAsNumpyArray, fmt="%s", delimiter=",", comments='', header=headerString)
    return reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList, is_sticking_coefficientList

def getReactionParametersFromCanteraReactionObjectsList(reactionObjectList):
    reactionIDsList = []
    reactionTypesList = []
    reactionEquationsList = []
    ArrheniusParametersList =[] #List of numpy arrays with A,b,E for each reaction.
    coverageDependencesList =[]
    coverageDependencesSpeciesList = []
    is_sticking_coefficientList =[]
    for reactionObject in reactionObjectList:
        reactionIDsList.append(reactionObject.ID)
        try:
            if int(reactionObject.reaction_type) == int(20):
                reactionTypesList.append("surface_reaction")
            if int(reactionObject.reaction_type) == int(1):
                reactionTypesList.append("reaction")
            if int(reactionObject.reaction_type) == int(2):
                reactionTypesList.append("three_body_reaction")
            if int(reactionObject.reaction_type) == int(4):
                reactionTypesList.append("falloff_reaction")
        except:
            reactionTypesList.append("None")
        #Now need to modify equation based on whether it's irreversible or not.
        reactionEquationsList.append( reactionObject.equation)
        A = float(reactionObject.rate.pre_exponential_factor)
        b = float(reactionObject.rate.temperature_exponent)
        E = float(reactionObject.rate.activation_energy)/1000 #Annoyingly, Cantera puts Energies out in J/kmol.
        singleReactionArrheniusParametersArray = np.array([A,b,E])
        ArrheniusParametersList.append(singleReactionArrheniusParametersArray)
        #Now check for coverage parameters.
        try:
            if len(reactionObject.coverage_deps) > 0:
                coverage_deps = True
            else:
                coverage_deps = False
        except:
                coverage_deps = False
        if coverage_deps == True:
            singleReactionCoverageDependencesDict = reactionObject.coverage_deps
            temporaryCoverageDependenceSpeciesList = list(reactionObject.coverage_deps.keys())
            if len(temporaryCoverageDependenceSpeciesList) > 0:
                coverageDependenceExists = True
                coverageDependenceSpecies = temporaryCoverageDependenceSpeciesList[0]
                a = reactionObject.coverage_deps[coverageDependenceSpecies][0]
                m = reactionObject.coverage_deps[coverageDependenceSpecies][1]
                e = float(reactionObject.coverage_deps[coverageDependenceSpecies][2])/1000 #Annoyingly, Cantera puts Energies out in J/kmol.   
        else:#There is no coverage dependence, so will put 'NaN')
             a = float('nan')
             m = float('nan')
             e = float('nan')
             coverageDependenceSpecies = 'None'
        coverageDependencesList.append(np.array([a,m,e]))
        coverageDependencesSpeciesList.append(coverageDependenceSpecies)
        is_sticking_coefficientList.append(reactionObject.is_sticking_coefficient)
    reactionIDsList = reactionIDsList
    reactionTypesList = reactionTypesList
    ArrheniusParametersArray = np.array(ArrheniusParametersList)
    coverageDependencesArray = np.array(coverageDependencesList)
    coverageDependencesSpeciesList = coverageDependencesSpeciesList
    is_sticking_coefficientList = is_sticking_coefficientList
    return reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList, is_sticking_coefficientList
    
def make_reaction_cti_string(individualreactions_parameters_array):
    #input should be a an array of strings: 
    #reactionID	canteraReactionType	reactionEquation	A	b	E	a	m	e	coverageDependenceSpecies	 is_sticking_coefficient
    # 1	20	H2 + 2 PT(S) => 2 H(S)	0.046	0	0	0	-1	0	PT(S)	TRUE
    #canteraReactionType is typically "surface_reaction" or "reaction". THe falloff and three_body reactions are not yet supported at this time.
    reaction_cti_string = None #This is to help diagnose errors.
    reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
    individualReactionTypeReceived = str(individualreactions_parameters_array[1])
    reactionEquation = str(individualreactions_parameters_array[2])
    A = str(individualreactions_parameters_array[3])
    b = str(individualreactions_parameters_array[4])
    E = str(individualreactions_parameters_array[5])
    a = str(individualreactions_parameters_array[6])
    m = str(individualreactions_parameters_array[7])
    e = str(individualreactions_parameters_array[8])
    coverageDependenceSpecies = str(individualreactions_parameters_array[9])
    is_sticking = str(individualreactions_parameters_array[10])
    if is_sticking.capitalize() == "False": #considered using distutils.util.strtobool(is_sticking) but decided that would slow the program down.
        ArrheniusString = "Arrhenius"
    if is_sticking.capitalize() == "True":
        ArrheniusString = "stick"
        
    #Now make the reaction_cti strings...
    if individualReactionTypeReceived.lower() == "reaction":
        reaction_cti_string = '''reaction('{0}',
        {1}({2}, {3}, {4}])'''.format(reactionEquation, ArrheniusString, A,b,E)
    if individualReactionTypeReceived.lower() == "surface_reaction":
        if coverageDependenceSpecies.capitalize() == "None": #FIXME: Not working yet (because of Cantera side, not because of this script.)
            reaction_cti_string= '''surface_reaction("{0}",
            {1}({2}, {3}, {4}))'''.format(reactionEquation,ArrheniusString,A,b,E) 
                #Like the below example.
                #        R3 = ct.InterfaceReaction.fromCti('''surface_reaction('O + H2 <=> H + OH',
                #        [3.87e1, 2.7, 2.619184e7])''')                      
        if coverageDependenceSpecies.capitalize() != "None": #FIXME: Not working yet (because of Cantera side, not because of this script.)
            reaction_cti_string = '''surface_reaction("{0}",
            {1}({2}, {3}, {4},
            coverage = ['{5}', {6}, {7}, {8}]))'''.format(reactionEquation,ArrheniusString,A,b,E,coverageDependenceSpecies,a,m,e)
                #Like the below example.
                #        R5 = ct.InterfaceReaction.fromCti('''surface_reaction( "CH4 + PT(S) + O(S) => CH3(S) + OH(S)",
                #                  Arrhenius(5.0e18, 0.7, 2000.0,
                #                            coverage = ['O(S)', 0.0, 0.0, 8000]))''')
    
    #now modfiy the reaction in the actual model:
    return reaction_cti_string


def create_full_cti(model_name, reactions_parameters_array = [], cti_top_info_string = None, write_cti_to_file = False):
    #It's convenient to use only the modelname. This then REQUIRES the reaction parameters and top info to
    #already be inside files with names of model_name+"_input_reactions_parameters.csv" and model_name+"_cti_top_info.cti"
    if cti_top_info_string == None:
       cti_top_info_filename = model_name + "_cti_top_info.cti"
       with open(cti_top_info_filename, "r") as cti_top_info_file:
           cti_top_info_string = cti_top_info_file.read() 
    #Normally, somebody will have to fill out the reaction parameters file ahead of time.
    #This time, it was made through the following line:
    #reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList, is_sticking_coefficientList = exportReactionParametersFromFile("ceO2_cti_full_existing.cti", "ceO2_input_reactions_parameters.csv")
    if len(reactions_parameters_array) == len([]):
       reaction_parameters_filename = model_name+"_input_reactions_parameters.csv"
       reactions_parameters_array = np.genfromtxt(reaction_parameters_filename, delimiter=",", skip_header=1, dtype="str")
       
    reactionsDataHeader = '\n\n\
#------------------------------------------------------------------------------- \n\
#  Reaction data \n\
#------------------------------------------------------------------------------- \n\
                           \n'
    #Now make the reactions string.
    cti_reactions_string = '\n' #Good to start with a new line.
    for element in reactions_parameters_array:
        cti_reactions_string += make_reaction_cti_string(element) + "\n" #need line breaks between each reaction_cti_string
    
    #Now generate the full cti.  
    cti_string = cti_top_info_string+reactionsDataHeader+cti_reactions_string
    if write_cti_to_file == True:
        cti_filename = model_name + "_cti_full.cti" 
        with open(cti_filename, 'w') as cti_file:
            cti_file.write(cti_string)
    return cti_string

def findModifiableParameterIndices(reactions_parameters_array):
    reactions_parameters_array = np.atleast_2d(reactions_parameters_array) #If for some reason a person is using only 1 reaction, we still need to make it 2D.
    listOfModifiableParameterIndices = []
    listOfOriginalValues = []
    for reactionIndex, individualreactions_parameters_array in enumerate(reactions_parameters_array):
        #This is the format:
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        individualReactionTypeReceived = str(individualreactions_parameters_array[1])
        reactionEquation = str(individualreactions_parameters_array[2])
        A = str(individualreactions_parameters_array[3])
        b = str(individualreactions_parameters_array[4])
        E = str(individualreactions_parameters_array[5])
        a = str(individualreactions_parameters_array[6])
        m = str(individualreactions_parameters_array[7])
        e = str(individualreactions_parameters_array[8])
        coverageDependenceSpecies = str(individualreactions_parameters_array[9])
        is_sticking = str(individualreactions_parameters_array[10])

        #For all reactions, we can modify the A,b,E parameters.
        listOfModifiableParameterIndices = listOfModifiableParameterIndices + [ [reactionIndex,3],[reactionIndex,4],[reactionIndex,5] ]
        #Now check for other parameters...
        if individualReactionTypeReceived == 'surface_reaction': #This can include adsorption, desorption, and surface reactions.
            if coverageDependenceSpecies != 'None': #If this is not none, then that means the a, m, and e can also be modified.
                listOfModifiableParameterIndices = listOfModifiableParameterIndices + [ [reactionIndex,6],[reactionIndex,7],[reactionIndex,8] ]
                #Currently, we cannot modify the coverageDependenceSpecies.
        #Other reaction types are not yet supported, but can be added in the future.                

    return listOfModifiableParameterIndices

def getAllModifiableReactionParametersValues(reactions_parameters_array, listOfModifiableParameterIndices=[]):
    if listOfModifiableParameterIndices == []: #If this has not been provided, we can get it.
        listOfModifiableParameterIndices = findModifiableParameterIndices(reactions_parameters_array)
    listOfOriginalValues = [] #Make the list that will be populated and returned.
    for modifiableParameterIndex in range(len(listOfModifiableParameterIndices)):
        reactionIndex = listOfModifiableParameterIndices[modifiableParameterIndex][0] #This is like the reaction ID, only it starts at 0.
        individualReactionParameterIndex = listOfModifiableParameterIndices[modifiableParameterIndex][1]#This says which column, basically whether it is A, b, E, etc.
        originalValue = reactions_parameters_array[reactionIndex][individualReactionParameterIndex] #Note that we use strings in the reaction parameters array.
        listOfOriginalValues.append(originalValue)
    return listOfOriginalValues

def modifyAllAdjustableReactionParametersArrayValues(reactions_parameters_array, newParametersList): #The first argument is a 2D numpy array. The second is a 1D list.
    reactions_parameters_array = np.atleast_2d(reactions_parameters_array) #If for some reason a person is using only 1 reaction, we still need to make it 2D.
    listOfModifiableParameterIndices = findModifiableParameterIndices(reactions_parameters_array)
    if len(listOfModifiableParameterIndices) != len(newParametersList):
        print("WARNING: modifyAllAdjustablereactions_parameters_arrayValues has been called with a newParametersList that is not of matching length.")
    #Now to make the modified parameters array...
    modified_reactions_parameters_array = copy.deepcopy(reactions_parameters_array)
    for newParameterIndex in range(len(newParametersList)):
        newValue = newParametersList[newParameterIndex]
        reactionIndex = listOfModifiableParameterIndices[newParameterIndex][0] #This is like the reaction ID, only it starts at 0.
        individualReactionParameterIndex = listOfModifiableParameterIndices[newParameterIndex][1]#This says which column, basically whether it is A, b, E, etc.
        modified_reactions_parameters_array[reactionIndex][individualReactionParameterIndex] = str(newValue) #Note that we use strings in the reaction parameters array.
    return modified_reactions_parameters_array

    
'''
#DEPRECATED BECAUSE NOW WE USE CANTERA CODE OR CANTERA OBJECTS
def makeDictfromXMLFile(FileName): #Filename is a string, can include path.
    with open("methane_pox_on_pt.xml") as xmlfile:
        xmldict = xmltodict.parse(xmlfile.read())
    return xmldict
#DEPRECATED BECAUSE NOW WE USE CANTERA CODE OR CANTERA OBJECTS
def makeReactionObjectsListFromCTMLasDict(CTMLasDict):
    reactionObjectsList = CTMLasDict["ctml"]["reactionData"]['reaction']
    return reactionObjectsList

#DEPRECATED BECAUSE NOW WE USE CANTERA CODE OR CANTERA OBJECTS  
    #NOte: does not have is_sticking_coefficientList and would need that to be useful.
def getReactionParametersFromReactionObjectsList(reactionObjectList):
    reactionIDsList = []
    reactionTypesList = []
    reactionEquationsList = []
    ArrheniusParametersList =[] #List of numpy arrays with A,b,E for each reaction.
    coverageDependencesList =[]
    coverageDependencesSpeciesList = []
    for reactionObject in reactionObjectList:
        reactionIDsList.append(reactionObject["@id"])
        try:
            reactionTypesList.append(reactionObject["@type"])
        except:
            reactionTypesList.append("None")
        #Now need to modify equation based on whether it's irreversible or not.
        reactionEquation = reactionObject["equation"]
        if reactionObject["@reversible"]=="no":
            reactionEquation = reactionEquation.replace("=]","=>")
        if reactionObject["@reversible"]=="yes":
            reactionEquation = reactionEquation.replace("=]","<=>")
        reactionEquationsList.append(reactionEquation)
        A = float(reactionObject['rateCoeff']['Arrhenius']['A'])
        b = float(reactionObject['rateCoeff']['Arrhenius']['b'])
        if reactionObject['rateCoeff']['Arrhenius']['E']['@units']=="J/mol":
            E = float(reactionObject['rateCoeff']['Arrhenius']['E']['#text']) #the xml is written in a way that this is necessary.
        else:
            E = 0.0
            print("Error: Currently, ony J/mol is supported for units of E. Fix parameters for this reaction: " + reactionObject['equation'])
        singleReactionArrheniusParametersArray = np.array([A,b,E])
        ArrheniusParametersList.append(singleReactionArrheniusParametersArray)
        #Now check for coverage parameters.
        if "coverage" in reactionObject['rateCoeff']['Arrhenius']:
            singleReactionCoverageDependencesDict = reactionObject['rateCoeff']['Arrhenius']['coverage']
            speciesDependence = singleReactionCoverageDependencesDict['@species']
            a = float(singleReactionCoverageDependencesDict['a'])
            m = float(singleReactionCoverageDependencesDict['m'])
            if singleReactionCoverageDependencesDict['e']['@units']=="J/mol":
                e = float(singleReactionCoverageDependencesDict['e']['#text']) #the xml is written in a way that this is necessary.
            else:
                e = 0.0
                print("Error: Currently, ony J/mol is supported for units of e. Fix coverage parameters for this reaction: " + reactionObject['equation'])
        else:#There is no coverage dependence, so will put 'NaN')
             a = float('nan')
             m = float('nan')
             e = float('nan')
             speciesDependence = 'None'
        coverageDependencesList.append(np.array([a,m,e]))
        coverageDependencesSpeciesList.append(speciesDependence)
    reactionIDsList = reactionIDsList
    reactionTypesList = reactionTypesList
    ArrheniusParametersArray = np.array(ArrheniusParametersList)
    coverageDependencesArray = np.array(coverageDependencesList)
    coverageDependencesSpeciesList = coverageDependencesSpeciesList
    return reactionIDsList, reactionTypesList, reactionEquationsList, ArrheniusParametersArray, coverageDependencesArray, coverageDependencesSpeciesList
'''




def ArrheniusParameterAddedToInOnePhase(canteraModule, canteraPhaseObject, ArrheniusParametersOperandsArray, byProvidedReactionID = True):
    #ArrheniusParametersMultipliersArray should have the same form as the reactions_parameters_array. 
    reactions_parameters_array = ArrheniusParametersOperandsArray #just using same local variable for convenience.
    ct = canteraModule
    for individualreactions_parameters_array in reactions_parameters_array:      
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        individualReactionTypeReceived = str(individualreactions_parameters_array[1])
        reactionEquation = str(individualreactions_parameters_array[2])
        A_operand = float(individualreactions_parameters_array[3])
        b_operand = float(individualreactions_parameters_array[4])
        E_operand = float(individualreactions_parameters_array[5])
        a_operand = float(individualreactions_parameters_array[6])
        m_operand = float(individualreactions_parameters_array[7])
        e_operand = float(individualreactions_parameters_array[8])
        coverageDependenceSpecies = str(individualreactions_parameters_array[9])

        if byProvidedReactionID == False: #optional.... 
                #byProvidedReactionID is set to False, but this only works if the cantera model has the equation written exactly the same.
                reactionID = canteraPhaseObject.reaction_equations().index(reactionEquation) #don't add 1 because reaction index is not python array indexing.

        #First get the existing object and its values so that they can be multiplied.
        existingReactionObject = canteraPhaseObject.reactions()[int(reactionID)]
        existingA = float(existingReactionObject.rate.pre_exponential_factor)
        existingb = float(existingReactionObject.rate.temperature_exponent)
        existingE = float(existingReactionObject.rate.activation_energy)
        temporaryCoverageDependenceSpeciesList = list(existingReactionObject.coverage_deps.keys())
        if len(temporaryCoverageDependenceSpeciesList) > 0:
            coverageDependenceExists = True
            coverageDependenceSpecies = temporaryCoverageDependenceSpeciesList[0]
            existinga = existingReactionObject.coverage_deps[coverageDependenceSpecies][0]
            existingm = existingReactionObject.coverage_deps[coverageDependenceSpecies][1]
            existinge = existingReactionObject.coverage_deps[coverageDependenceSpecies][2]
        existingReactionObject.rate = ct.Arrhenius(float(existingA+A_operand), float(existingb+b_operand), float(existingE+E_operand))
        if coverageDependenceExists == True:
            print("Warning: Coverage dependence modifiers not working yet")
            try:
                if a_operand+m_operand+e_operand != 0.0: #FIXME: Not working yet (because of Cantera side, not because of this script.)
                    tupleForCoverageDependence = (existinga+a_operand ,existingm+m_operand , existinge+e_operand)
                    existingReactionObject.coverage_deps[coverageDependenceSpecies]=tupleForCoverageDependence
            except:
                pass
        canteraPhaseObject.modify_reaction(int(reactionID), existingReactionObject)           
               
def ArrheniusParametersMultiplierInOnePhase(canteraModule, canteraPhaseObject, ArrheniusParametersMultipliersArray, byProvidedReactionID = True):
    #ArrheniusParametersMultipliersArray should have the same form as the reactions_parameters_array. 
    reactions_parameters_array = ArrheniusParametersMultipliersArray #just using same local variable for convenience.
    ct = canteraModule
    for individualreactions_parameters_array in reactions_parameters_array:      
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        individualReactionTypeReceived = str(individualreactions_parameters_array[1])
        reactionEquation = str(individualreactions_parameters_array[2])
        A_multiplier = float(individualreactions_parameters_array[3])
        b_multiplier = float(individualreactions_parameters_array[4])
        E_multiplier = float(individualreactions_parameters_array[5])
        a_multiplier = float(individualreactions_parameters_array[6])
        m_multiplier = float(individualreactions_parameters_array[7])
        e_multiplier = float(individualreactions_parameters_array[8])
        coverageDependenceSpecies = str(individualreactions_parameters_array[9])

        if byProvidedReactionID == False: #optional.... 
                #byProvidedReactionID is set to False, but this only works if the cantera model has the equation written exactly the same.
                reactionID = canteraPhaseObject.reaction_equations().index(reactionEquation) #don't add 1 because reaction index is not python array indexing.

        #First get the existing object and its values so that they can be multiplied.
        existingReactionObject = canteraPhaseObject.reactions()[int(reactionID)]
        existingA = float(existingReactionObject.rate.pre_exponential_factor)
        existingb = float(existingReactionObject.rate.temperature_exponent)
        existingE = float(existingReactionObject.rate.activation_energy)
        temporaryCoverageDependenceSpeciesList = list(existingReactionObject.coverage_deps.keys())
        if len(temporaryCoverageDependenceSpeciesList) > 0:
            coverageDependenceExists = True
            coverageDependenceSpecies = temporaryCoverageDependenceSpeciesList[0]
            existinga = existingReactionObject.coverage_deps[coverageDependenceSpecies][0]
            existingm = existingReactionObject.coverage_deps[coverageDependenceSpecies][1]
            existinge = existingReactionObject.coverage_deps[coverageDependenceSpecies][2]
        existingReactionObject.rate = ct.Arrhenius(float(existingA*A_multiplier), float(existingb*b_multiplier), float(existingE*E_multiplier))
        if coverageDependenceExists == True:
            print("Warning: Coverage dependence modifiers not working yet")
            try:
                if a_multiplier*m_multiplier*e_multiplier != 1.0: #FIXME: Not working yet (because of Cantera side, not because of this script.)
                    tupleForCoverageDependence = (existinga*a_multiplier ,existingm*m_multiplier , existinge*e_multiplier)
                    existingReactionObject.coverage_deps[coverageDependenceSpecies]=tupleForCoverageDependence
            except:
                pass
        canteraPhaseObject.modify_reaction(int(reactionID), existingReactionObject)           

def populatePiecewiseCoverageDependence(simulation_settings_module, original_reactions_parameters_array, species_name, kineticParameterName, piecewise_coverage_intervals, modifiers_array):
    try:
        len(simulation_settings_module.piecewise_coverage_dependences)
    except: #if it has no length, then dictionary has not been created yet.
        simulation_settings_module.piecewise_coverage_dependences = {}
    try:
        len(simulation_settings_module.piecewise_coverage_dependences[species_name])
    except: #if it has no length, then dictionary has not been created yet.
        simulation_settings_module.piecewise_coverage_dependences[species_name] = {}
        simulation_settings_module.piecewise_coverage_dependences[species_name]['piecewise_coverage_intervals'] = piecewise_coverage_intervals
        simulation_settings_module.piecewise_coverage_dependences[species_name]['piecewise_kinetic_parameter_modifier_arrays']={}
    simulation_settings_module.original_reactions_parameters_array = original_reactions_parameters_array
    simulation_settings_module.piecewise_coverage_dependences[species_name]['piecewise_kinetic_parameter_modifier_arrays'][kineticParameterName] = np.array(modifiers_array)
    return simulation_settings_module.piecewise_coverage_dependences #Normally don't use the return. Normally just use it to populate the variables in the simulation_settings_module.

def getInterpolatedModifiersArray(existing_values,piecewise_coverage_dependences, species_name, species_coverage, kineticParameterKey, reaction_parameters_array_kinetic_parameter_index):
    all_coverages_modifiers_array = np.array(piecewise_coverage_dependences[species_name]['piecewise_kinetic_parameter_modifier_arrays'][kineticParameterKey],dtype="float") #This is for every coverage dependent value for this species for this kinetic parameter (i.e., coverages of 0 to 1.0 in bins)
    if len(all_coverages_modifiers_array) != len(existing_values):
        print("You have piecewise_coverage_dependence set to True in the simulation settings. However, the coverage dependence array for kinetic parameter " + kineticParameterKey + " has a function of the coverage of" + species_name  + "does not match the length of the kinetic parameters array. These lengths must match to use this feature.")
    all_piecewise_coverage_intervals = np.array(piecewise_coverage_dependences[species_name]["piecewise_coverage_intervals"], dtype="float") #This is the list of coverages to interpolate between to find out which index to extract from.
    index_of_relevant_interval_low = np.searchsorted(all_piecewise_coverage_intervals, species_coverage, side="right") - 1 #Prefer to find lower coverage, but can only get right side of interval by searchsorted. (side="left doesn't work usefully for us).
    index_of_relevant_interval_high = index_of_relevant_interval_low + 1 

     #If the coverage value is less than or equal to the lowest coverage defined, we use the lowest index's value.
    if species_coverage <= min(all_piecewise_coverage_intervals):
        indexOfMin = np.where(all_piecewise_coverage_intervals == min(all_piecewise_coverage_intervals) )#We don't assume the user gave an ascending list.
        interpolated_modifiers_array = np.array(all_coverages_modifiers_array[:,indexOfMin]).flatten() #What comes out is nested and so we flatten it.
     #If the coverage value is greater than or equal to the highest coverage defined, we use the highest index's value.
    elif species_coverage >= max(all_piecewise_coverage_intervals):
        indexOfMax = np.where(all_piecewise_coverage_intervals == max(all_piecewise_coverage_intervals) )#We don't assume the user gave an ascending list.
        interpolated_modifiers_array = np.array(all_coverages_modifiers_array[:,indexOfMax]).flatten()  #What comes out is nested and so we flatten it.
    else:
        #Normal cases... we will interpolate to find the right modifier value between the nearest coverage below and above (denoted "low" and "high").
        modifiers_array_low = np.array(all_coverages_modifiers_array[:,index_of_relevant_interval_low],dtype="float") # The ":" is for slicing across all reactions.
        modifiers_array_high = np.array(all_coverages_modifiers_array[:,index_of_relevant_interval_high],dtype="float") # The ":" is for slicing across all reactions.
        piecewise_interval_low = np.array(all_piecewise_coverage_intervals[index_of_relevant_interval_low],dtype="float")
        piecewise_interval_high = np.array(all_piecewise_coverage_intervals[index_of_relevant_interval_high],dtype="float")
        slopes_array = (modifiers_array_high-modifiers_array_low)/(piecewise_interval_high-piecewise_interval_low)
        intercepts_array = modifiers_array_high - 1.0*slopes_array*piecewise_interval_high #b = y - mx
        #now from interpolation y = mx+b means  modifiers_array = slopes_array*species_coverage+intercepts_array
        interpolated_modifiers_array = slopes_array*species_coverage+intercepts_array
    return interpolated_modifiers_array

def calculatePiecewiseCoverageDependentModifiedParametersArray(settingsModuleOrObject, species_names, species_coverages): #To use this feature, a module or object must be passed which has all of the arrays inside with the correct names. Typically, simulation_settings_module would be used.
    #The data structure here is important to understand know how this function works:
    #The settingsModuleOrObject is expected to have the original_reactions_parameters_array.
    #The settingsModuleOrObject is also expected to have the piecewise_coverage_dependences in it, which is nested. It has a structure of dictionaries inside such that arrays are pulled out like this:  
    #    piecewise_coverage_dependences[species_name][kineticParameterName] 
    original_reactions_parameters_array = settingsModuleOrObject.original_reactions_parameters_array #This feature requires the settingsModuleOrObject to already have the reactions_parameters_array assigned to it.
    modified_reactions_parameters_array = np.array(copy.deepcopy(original_reactions_parameters_array), dtype="str") #Normally the original shoudl already be a str based np array, but it could be a nested list. We need to make it a numpy array for slicing.
    piecewise_coverage_dependences = settingsModuleOrObject.piecewise_coverage_dependences
    #TODO: consider adding other parameters. For now, only A_values and E_values are supported, though of course others can be supported. In the case of A, the piecewise coverage dependence is a factor 10^modifier.
    for species_name in piecewise_coverage_dependences.keys(): #Looping across species for which there are modifications terms. #the piecewise_coverage_dependance object is a dictionary with species as keys. Within that, are dictionaries for each species' contributions to the coverage dependences. Note that this may not represent ALL species.
        species_index = species_names.index(species_name) #This is only the index of the species in the species_coverages list. The piecewise coverage dependences dictionary may have fewer species.
        species_coverage = float(species_coverages[species_index])
        for kineticParameterKey in piecewise_coverage_dependences[species_name]['piecewise_kinetic_parameter_modifier_arrays']: #looping across kinetic parameters for which there are coverage dependent modifier terms. #There could be all of the kinetic parameter keys present, or only some subset.
            if kineticParameterKey == "A":
                reaction_parameters_array_kinetic_parameter_index = 3 #This is the 'column' in the reaction parameters array where the kinetic parameter is kept.             

                #NOTE: we take the modified_reactions_parameters_array because this way modifications can stack.
                existing_parameter_values = np.array(modified_reactions_parameters_array[:,reaction_parameters_array_kinetic_parameter_index], dtype="float") #NOTE: we modify the current_reactions_parameters_array because this way modifications can stack.
                modifiers_array= getInterpolatedModifiersArray(modified_reactions_parameters_array,piecewise_coverage_dependences, species_name, species_coverage, kineticParameterKey, reaction_parameters_array_kinetic_parameter_index)
    
                #now to apply the modifiers...
                modified_values = existing_parameter_values*(10**modifiers_array) #In the case of A, we apply a coefficient of 10^modifiers_array_value.
                #updated the values in reaction parameters array, but make them into a string.
                modified_reactions_parameters_array[:,reaction_parameters_array_kinetic_parameter_index] = np.array(modified_values, dtype="str")
            if kineticParameterKey == "E":
                reaction_parameters_array_kinetic_parameter_index = 5  #This is the 'column' in the reaction parameters array where the kinetic parameter is kept.
                
                #NOTE: we take the modified_reactions_parameters_array because this way modifications can stack.
                existing_parameter_values = np.array(modified_reactions_parameters_array[:,reaction_parameters_array_kinetic_parameter_index], dtype="float") #NOTE: we modify the current_reactions_parameters_array because this way modifications can stack.
                modifiers_array= getInterpolatedModifiersArray(modified_reactions_parameters_array,piecewise_coverage_dependences, species_name, species_coverage, kineticParameterKey, reaction_parameters_array_kinetic_parameter_index)
                
                modified_values = existing_parameter_values+modifiers_array #For E, we add the modifiers.
                #updated the values in reaction parameters array, but make them into a string.
                modified_reactions_parameters_array[:,reaction_parameters_array_kinetic_parameter_index] = np.array(modified_values, dtype="str")    
    return modified_reactions_parameters_array 
        

def multiplyReactionsInOnePhase(canteraModule, canteraPhaseObject, reactions_parameters_array, rateMultipliersArray, byProvidedReactionID = True):
    ct = canteraModule
    for inputReactionIndex,individualreactions_parameters_array in enumerate(reactions_parameters_array):    
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        reactionEquation = str(individualreactions_parameters_array[2])
        if byProvidedReactionID == False: #optional.... 
            #byProvidedReactionID is set to False, but this only works if the cantera model has the equation written exactly the same.
            reactionID = canteraPhaseObject.reaction_equations().index(reactionEquation) #don't add 1 because reaction index is not python array indexing.
        canteraPhaseObject.set_multiplier(rateMultipliersArray[inputReactionIndex],int(reactionID)) #note that we are iterating across each individual reaction, but that the reactionID may not match the inputReactionIndex
    
def modifyReactionsInOnePhase(canteraPhaseObject, reactions_parameters_array, ArrheniusOnly = True, byProvidedReactionID = True):
    #if byProvidedReactionID is set to false, then the reaction ID is "looked up" from the reactions list in the canteraPhaseObject.
    #The "ArrheniusOnly = True" flag is because cantera currently does not support non-Arrhenius modify_reaction for surface reactions.
    #The "multiplierOnly = True" flag just modifies the existing pre-exponential by a specified factor.
    
    import cantera as ct
    #First classify what type of phase we have.
    if(str(type(canteraPhaseObject)))== "<class 'cantera.composite.Interface'>":
        phaseType = "Interface"
        inputReactionTypeNeeded = "surface_reaction"
        #outputReactionObjectType = "InterfaceReaction" 
    if(str(type(canteraPhaseObject)))== "<class 'cantera.composite.Solution'>":
        phaseType = "Solution"
        inputReactionTypeNeeded = "reaction"
        #outputReactionObjectType = "Reaction" 

    #Now add any reactions that match for this phase.
    for individualreactions_parameters_array in reactions_parameters_array:      
        reactionID = str(int(individualreactions_parameters_array[0])-1) #we need to subtract one for how cantera expects the reaction IDs when modifying.
        individualReactionTypeReceived = str(individualreactions_parameters_array[1])
        reactionEquation = str(individualreactions_parameters_array[2])
        A = str(individualreactions_parameters_array[3])
        b = str(individualreactions_parameters_array[4])
        E = str(individualreactions_parameters_array[5])
        a = str(individualreactions_parameters_array[6])
        m = str(individualreactions_parameters_array[7])
        e = str(individualreactions_parameters_array[8])
        coverageDependenceSpecies = str(individualreactions_parameters_array[9])

        if byProvidedReactionID == False: #optional.... 
                #byProvidedReactionID is set to False, but this only works if the cantera model has the equation written exactly the same.
                reactionID = canteraPhaseObject.reaction_equations().index(reactionEquation) #don't add 1 because reaction index is not python array indexing.

        if ArrheniusOnly == True:
            if individualReactionTypeReceived == inputReactionTypeNeeded: #ignore any case that does not match what's needed for this phase.
                existingReactionObject = canteraPhaseObject.reactions()[int(reactionID)]
                existingReactionObject.rate = ct.Arrhenius(float(A), float(b), float(E)*1000) #If modifying cantera Arrhenius objects, need to convert from J/mol to J/kmol.
                canteraPhaseObject.modify_reaction(int(reactionID), existingReactionObject)     
        if ArrheniusOnly == False:
            if individualReactionTypeReceived == inputReactionTypeNeeded: #ignore any case that does not match what's needed for this phase.
                if phaseType == "Solution":
                    cti_string = '''reaction('{0}', [{1}, {2}, {3}])'''.format(reactionEquation, A,b,E)
                    modifiedReactionObject = ct.Reaction.fromCti(cti_string)
                if phaseType == "Interface":
                    if coverageDependenceSpecies == "None": #FIXME: Not working yet (because of Cantera side, not because of this script.)
                        cti_string= '''surface_reaction("{0}",
                            [{1}, {2}, {3}])'''.format(reactionEquation,A,b,E) 
                        modifiedReactionObject = ct.InterfaceReaction.fromCti(cti_string)
                            #Like the below example.
                            #        R3 = ct.InterfaceReaction.fromCti('''surface_reaction('O + H2 <=> H + OH',
                            #        [3.87e1, 2.7, 2.619184e7])''')                      
                    if coverageDependenceSpecies != "None": #FIXME: Not working yet (because of Cantera side, not because of this script.)
                        print("Warning: Coverage dependence modifiers not working yet")
                        cti_string = '''surface_reaction("{0}",Arrhenius({1}, {2}, {3},
                                            coverage = ['{4}', {5}, {6}, {7}]))'''.format(reactionEquation,A,b,E,coverageDependenceSpecies,a,m,e)
                        modifiedReactionObject = ct.InterfaceReaction.fromCti(cti_string)                    
                            #Like the below example.
                            #        R5 = ct.InterfaceReaction.fromCti('''surface_reaction( "CH4 + PT(S) + O(S) => CH3(S) + OH(S)",
                            #                  Arrhenius(5.0e18, 0.7, 2000.0,
                            #                            coverage = ['O(S)', 0.0, 0.0, 8000]))''')
               
                #now modfiy the reaction in the actual model:
                canteraPhaseObject.modify_reaction(int(reactionID),modifiedReactionObject) 


def main():
    pass


if __name__ == '__main__':
    main()


# Type 'pytest' in the current directory to run tests.
# EAW 2020/01/17
import plotting_functions
from plotting_functions import plotting_functions
import UserInput_ODE_KIN_BAYES_SG_EW as UserInput
import run_me
from run_me import ip

def test_mumpce_plots():
    plot_object = plotting_functions()
    assert plot_object.mumpce_plots(model_parameter_info = UserInput.model_parameter_info, active_parameters = UserInput.active_parameters, pairs_of_parameter_indices = UserInput.pairs_of_parameter_indices, posterior_mu_vector = UserInput.posterior_mu_vector, posterior_cov_matrix = UserInput.posterior_cov_matrix, prior_mu_vector = UserInput.prior_mu_vector, prior_cov_matrix = UserInput.prior_cov_matrix, contour_settings_custom = UserInput.contour_settings_custom) == None
    assert exec(open("mumpce_custom_plotting_example.py").read()) == None

def MH_then_scatterplot():
    parseUserInputParameters()
    ip_object = ip()
    [evidence, info_gain, samples, rate_tot_array, logP] = ip_object.MetropolisHastings()
    plotting_object = plotting_functions(samples=samples)
    assert plotting_object.seaborn_scatterplot_matrix() == None



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  5 16:10:05 2019

@author: fvs
"""
import sys
sys.path.insert(0, '/mumpce/')
import mumpce.Project as mumpceProject
import mumpce.solution as mumpceSolution
import numpy as np
import matplotlib
from matplotlib import cm #EAW 2020/01/07
#matplotlib.use('Agg') #EAW 2020/01/07

'''
THIS FILE IS TO DEMONSTRATE CUSTOM PLOTTING WITH MUMPCE SOLUTION OBJECTS
The method for using the custom plotting features is to make a dictionary named contour_settings_custom
and to pass this dictionary as an optional argument into either plot_pdfs or _single_pdf_plot
which are functions in the Project class from Project.py of mumpce.

#Only one or more key:item pair needs to be added into contour_settings_custom to turn it on.
The item added does not even need to be a real option.
#So for example:; contour_settings_custom['on'] = True #would turn it on.

The options are:
contour_settings_custom['zoom_std_devs'] = 2.5 #how zoomed in the image is.
contour_settings_custom['center_on'] = 'all' #can be 'all', 'prior' or 'posterior'. 
contour_settings_custom['fontsize'] #sets the fontsize for everything except the colorbars. Can be an integer or the word 'auto', or the word "None". Should change space_between_subplots if fontsize is changed. 
contour_settings_custom['space_between_subplots'] #Typically a value between 0.20 and 5.0. Set to 4.0 by default. Should be changed when font size is changed. Fontsize 'auto' tends to make small fonts which needs smaller values like 0.20.
contour_settings_custom["colorbars"] = True #can be true or false.
contour_settings_custom['num_pts_per_axis'] = 500 #This sets the resolution of the contours.
contour_settings_custom["cmap_levels"] = 4   #This is the number of contours.
contour_settings_custom['cmap_prior'] #define a color map directly
contour_settings_custom['cmap_posterior']  #define a color map directly
contour_settings_custom['colormap_prior_customized'] #accepts a string (matplotlib colormap names, like 'Greens') or a list of tuples with 0-to-1 and colornames to interpolate between. For example, the default right now is:  [(0,    '#00FFFF'),(1,    '#0000FF')]. The tuple could have 0, 0.7, and 1, for example. #colors can be obtained from: https://www.htmlcsscolor.com/hex/244162  
contour_settings_custom['colormap_posterior_customized'] #accepts a string (matplotlib colormap names, like 'Oranges') or a list of tuples with 0-to-1 and colornames to interpolate between. For example, the default right now is:  [(0,    '#FFFF00'),(1,    '#FF0000')]. The tuple could have 0, 0.7, and 1, for example. #colors can be obtained from: https://www.htmlcsscolor.com/hex/244162  
contour_settings_custom['contours_normalized'] = True #Makes the maximum 1, with the contours at 0.2, 0.4, 0.6., 0.8
contour_settings_custom['num_y_ticks'] #adusts number of y ticks (actually sets a maximum number of them). #num_y_ticks and num_x_ticks must be either a string ('auto') or an integer (such as 4, either without string or with integer casting like int('5')). This feature is recommended.  #Note that this is a *request* When it's not fulfilled exactly, the user can play with the number.
contour_settings_custom['num_x_ticks'] #adjusts number of x ticks (actually sets a maximum number of them). #num_y_ticks and num_x_ticks must be either a string ('auto') or an integer (such as 4, either without string or with integer casting like int('5')).This feature is recommended. #Note that this is a *request* When it's not fulfilled exactly, the user can play with the number.
contour_settings_custom['x_ticks'] #feed in an array of directly. Not recommended.
contour_settings_custom['y_ticks'] #feed in an array of directly. Not recommended.
contour_settings_custom['axis_limits'] #Feed in list of [x_min, x_max, y_min, y_max]. This is recommended for use.

'''

###THE BELOW LINES OF CODE SETUP THE DATA IN THE MUMPCE PROJECT OBJECT####
mumpceProjectObject = mumpceProject.Project()
mumpceProjectObject.model_parameter_info = np.array([{'parameter_number': 0, 'parameter_name': 'Parameter 0', 'parameter_value': 1.0},
 {'parameter_number': 1, 'parameter_name': 'Parameter 1', 'parameter_value': 1.0}, #The parameter values are not used in making these contour plots.
 {'parameter_number': 2, 'parameter_name': 'Parameter 2', 'parameter_value': 1.0},
 {'parameter_number': 3, 'parameter_name': 'Parameter 3', 'parameter_value': 1.0},
 {'parameter_number': 4, 'parameter_name': 'Parameter 4', 'parameter_value': 1.0},
 {'parameter_number': 5, 'parameter_name': 'Parameter 5', 'parameter_value': 1.0},
 {'parameter_number': 6, 'parameter_name': 'Parameter 6', 'parameter_value': 1.0}]) #This must be made into a numpy array.
mumpceProjectObject.active_parameters = np.array([0, 1, 2, 4, 6]) #this must be made into a numpy array.
#mumpceProjectObject.set_active_parameters = [0, 1, 2, 4, 6]
Posterior_mu_vector = np.array([-0.58888733,1.1200355, 0.00704044, -1.62385888,0.80439847]) #this must be made into a numpy array. #This will become solution.x
Posterior_cov_vector = np.array([[ 0.0148872,-0.01894579, -0.01047339,0.01325883,0.04734254],
 [-0.01894579,0.04284732, -0.00131389, -0.04801795, -0.04545703],
 [-0.01047339, -0.00131389,0.02343653,0.01588293, -0.05618226],
 [ 0.01325883, -0.04801795,0.01588293,0.08171972,0.00875017],
 [ 0.04734254, -0.04545703, -0.05618226,0.00875017,0.20669273]]) #This will become solution.cov. It does not need to be a numpy array, but we make it one for consistency.

Prior_mu_vector = np.array([-0.98888733,0.8200355, 0.01204044, -7.02385888,0.40439847])
Prior_cov_vector = 10*Posterior_cov_vector

mumpceSolutionsObject = mumpceSolution.Solution(Posterior_mu_vector, Posterior_cov_vector, initial_x=Prior_mu_vector, initial_covariance=Prior_cov_vector)
mumpceProjectObject.solution = mumpceSolutionsObject
mumpceProjectObject.pairsOfParameterIndices = [[0, 1], [1, 2],[3, 4]]


###THE BELOW LINES OF CODE MAKE EXAMPLE PLOTS WITH THE DEFAULT AND WITH THE NEW REVISED WAY####

#This makes the figures as originally programmed in mumpce, which assumes/requries the cov to be normalized to 1.
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices)

#I have expanded the code to allow more versatility an optional argument called contour_settings_custom.
#It does not assume/require that things be normalized to 1.


contour_settings_custom = {} 
contour_settings_custom['figure_name']='mumpce_plots_02'
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)


contour_settings_custom={}
contour_settings_custom['figure_name']='mumpce_plots_03'
contour_settings_custom['fontsize'] = 'auto'
contour_settings_custom['num_y_ticks'] = 'auto'
contour_settings_custom['num_x_ticks'] = 'auto'
contour_settings_custom['colorbars'] = True
contour_settings_custom['space_between_subplots'] = 0.20
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)

contour_settings_custom = {}
contour_settings_custom['figure_name']='mumpce_plots_04'
contour_settings_custom['fontsize'] = 'auto'
contour_settings_custom['num_y_ticks'] = 3 #Note that this is a *request* When it's not fulfilled exactly, the user can play with the number.
contour_settings_custom['num_x_ticks'] = 3 #Note that this is a *request* When it's not fulfilled exactly, the user can play with the number.
contour_settings_custom['colormap_posterior_customized'] = "Oranges"
contour_settings_custom['colormap_prior_customized'] = "Greens"
contour_settings_custom['contours_normalized'] = False
contour_settings_custom['center_on'] = 'prior'
contour_settings_custom['colorbars'] = True
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)

contour_settings_custom = {}
contour_settings_custom['figure_name']='mumpce_plots_05'
contour_settings_custom['space_between_subplots'] = 0.50
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)

contour_settings_custom = {}
contour_settings_custom['figure_name']='mumpce_plots_06'
contour_settings_custom['colorbars'] = True
contour_settings_custom['space_between_subplots'] = 0.50
contour_settings_custom['zoom_std_devs'] = 4.5 #These are in units of standard deviations of the prior.
mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)
